# -*- coding: utf-8 -*-

"""
Bucket related API.
"""

import typing as T


if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class BucketAPIMixin:
    """
    A mixin class that implements the bucket related methods.
    """



# -*- coding: utf-8 -*-

import warnings
import functools


def warn_deprecate(
    func_name: str,
    version: str,
    message: str,
):
    warnings.warn(f"{func_name!r} will be deprecated on {version}: {message}")


def deprecate_v1(version: str, message: str):
    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warn_deprecate(func.__name__, version, message)
            return func(*args, **kwargs)

        return wrapper

    return deco


def deprecate_v2(version: str, message: str):
    from decorator import decorator
    
    @decorator
    def deco(func, *args, **kwargs):
        warn_deprecate(func.__name__, version, message)
        return func(*args, **kwargs)

    return deco



# -*- coding: utf-8 -*-

"""
Smart open library integration.
"""

import typing as T

from .resolve_s3_client import resolve_s3_client
from ..aws import context
from ..compat import smart_open, compat
from ..type import MetadataType, TagType
from ..tag import encode_url_query

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class OpenerAPIMixin:
    """
    A mixin class that implements the file-object protocol.
    """
    def open(
        self: 'S3Path',
        mode: T.Optional[str] = "r",
        buffering: T.Optional[int] = -1,
        encoding: T.Optional[str] = None,
        errors: T.Optional[str] = None,
        newline: T.Optional[str] = None,
        closefd=True,
        opener=None,
        ignore_ext: bool = False,
        compression: T.Optional[str] = None,
        multipart_upload: bool = True,
        metadata: T.Optional[MetadataType] = None,
        tags: T.Optional[TagType] = None,
        transport_params: T.Optional[dict] = None,
        bsm: T.Optional['BotoSesManager'] = None,
    ):
        """
        Open S3Path as a file-liked object.

        :param mode: "r", "w", "rb", "wb"
        :param compression: whether do you want to compress the content
        :param multipart_upload: do you want to use multi-parts upload,
            by default it is True
        :param metadata: also put the user defined metadata dictionary
        :param tags: also put the tag dictionary
        :param bsm: optional ``boto_session_manager.BotoSesManager`` object

        :return: a file-like object that has ``read()`` and ``write()`` method.

        See https://github.com/RaRe-Technologies/smart_open for more info.

        .. versionadded:: 1.0.1

        .. versionchanged:: 1.2.1
        """
        s3_client = resolve_s3_client(context, bsm)
        if transport_params is None:
            transport_params = dict()
        transport_params["client"] = s3_client
        transport_params["multipart_upload"] = multipart_upload

        open_kwargs = dict(
            uri=self.uri,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
            transport_params=transport_params,
        )

        if compat.smart_open_version_major < 6:  # pragma: no cover
            open_kwargs["ignore_ext"] = ignore_ext
        if compat.smart_open_version_major >= 5 and compat.smart_open_version_major >= 1:  # pragma: no cover
            if compression is not None:
                open_kwargs["compression"] = compression

        # if any of additional parameters exists, we need additional handling
        if sum([metadata is not None, tags is not None]) > 0:
            kwargs = {}
            if metadata is not None:
                kwargs["Metadata"] = metadata
            if tags is not None:
                kwargs["Tagging"] = encode_url_query(tags)

            if multipart_upload:
                client_kwargs = {"S3.Client.create_multipart_upload": kwargs}
            else:
                client_kwargs = {"S3.Client.put_object": kwargs}
            if "client_kwargs" in transport_params: # pragma: no cover
                transport_params["client_kwargs"].update(client_kwargs)
            else:
                transport_params["client_kwargs"] = client_kwargs
        return smart_open.open(**open_kwargs)



# -*- coding: utf-8 -*-

"""
Metadata related API.
"""

import typing as T
import warnings
from datetime import datetime

from .resolve_s3_client import resolve_s3_client
from .filterable_property import FilterableProperty
from .. import utils, client as better_client
from ..aws import context

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


def alert_upper_case(metadata: dict):
    """
    Alert if there is upper case used in user defined metadata.

    Ref:

    - https://docs.aws.amazon.com/AmazonS3/latest/userguide/UsingMetadata.html#UserMetadata
    """
    for k, v in metadata.items():
        if k.lower() != k:
            msg = (
                f"based on this document "
                f"https://docs.aws.amazon.com/AmazonS3/latest/userguide/UsingMetadata.html#UserMetadata "
                f"Amazon will automatically convert user defined metadata to lower case, "
                f"but you have a key: {k!r} in metadata"
            )
            warnings.warn(msg, UserWarning)


class MetadataAPIMixin:
    """
    A mixin class that implements the metadata related methods.

    Note:

        1. only S3 object can have metadata.
        2. metadata is immutable.
        3. user metadata key is always lower case.
    """

    def head_object(
        self: "S3Path",
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> dict:
        """
        Call head_object() api, store metadata value.
        """
        s3_client = resolve_s3_client(context, bsm)
        dct = better_client.head_object(s3_client, self.bucket, self.key)
        self._meta = dct
        return dct

    def _get_meta_value(
        self: "S3Path",
        key: str,
        default: T.Any = None,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> T.Any:
        if self._meta is None:
            self.head_object(bsm=bsm)
        return self._meta.get(key, default)

    def _get_or_pull_meta_value(
        self: "S3Path",
        key: str,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> T.Any:
        """
        Note:

            This method is for those metadata fields that always exists.
        """
        value = self._get_meta_value(key, default=None, bsm=bsm)
        if value is None:
            self.head_object(bsm=bsm)
        return self._meta[key]

    @FilterableProperty
    def etag(self: "S3Path") -> str:
        """
        For small file, it is the md5 check sum. For large file, because it is
        created from multi part upload, it is the sum of md5 for each part and
        md5 of the sum.

        Ref: https://docs.aws.amazon.com/AmazonS3/latest/API/API_Object.html

        .. versionadded:: 1.0.1
        """
        return self._get_or_pull_meta_value(key="ETag")[1:-1]

    @FilterableProperty
    def last_modified_at(self: "S3Path") -> datetime:
        """
        Ref: https://docs.aws.amazon.com/AmazonS3/latest/API/API_Object.html

        .. versionadded:: 1.0.1
        """
        return self._get_or_pull_meta_value(key="LastModified")

    @FilterableProperty
    def size(self: "S3Path") -> int:
        """
        Ref: https://docs.aws.amazon.com/AmazonS3/latest/API/API_Object.html

        .. versionadded:: 1.0.1
        """
        return self._get_or_pull_meta_value(key="ContentLength")

    @property
    def size_for_human(self: "S3Path") -> str:
        """
        A human readable string version of the size.

        .. versionadded:: 1.0.1
        """
        return utils.repr_data_size(self.size)

    @FilterableProperty
    def version_id(self: "S3Path") -> int:
        """
        Only available if you turned on versioning for the bucket.

        Ref: https://docs.aws.amazon.com/AmazonS3/latest/API/API_Object.html

        .. versionadded:: 1.0.1
        """
        return self._get_meta_value(key="VersionId")

    @FilterableProperty
    def expire_at(self: "S3Path") -> datetime:
        """
        Only available if you turned on TTL

        Ref: https://docs.aws.amazon.com/AmazonS3/latest/API/API_Object.html

        .. versionadded:: 1.0.1
        """
        return self._get_meta_value(key="Expires")

    @property
    def metadata(self: "S3Path") -> dict:
        """
        Ref: https://docs.aws.amazon.com/AmazonS3/latest/API/API_Object.html

        .. versionadded:: 1.0.1
        """
        return self._get_or_pull_meta_value(key="Metadata")

    def clear_cache(self: "S3Path") -> None:
        """
        Clear all cache that stores metadata information.

        .. versionadded:: 1.0.1
        """
        self._meta = None

    @classmethod
    def _from_content_dict(cls: T.Type["S3Path"], bucket: str, dct: dict) -> "S3Path":
        """
        Construct S3Path object from the response["Content"] dictionary data.

        Example ``dct``::

            {
                'Key': 'string',
                'LastModified': datetime(2015, 1, 1),
                'ETag': 'string',
                'ChecksumAlgorithm': [
                    'CRC32'|'CRC32C'|'SHA1'|'SHA256',
                ],
                'Size': 123,
                'StorageClass': 'STANDARD'|'REDUCED_REDUNDANCY'|'GLACIER'|'STANDARD_IA'|'ONEZONE_IA'|'INTELLIGENT_TIERING'|'DEEP_ARCHIVE'|'OUTPOSTS'|'GLACIER_IR',
                'Owner': {
                    'DisplayName': 'string',
                    'ID': 'string'
                }
            }

        Ref:

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.list_objects
        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.list_objects_v2

        :return: a new S3Path object.
        """
        p = cls(bucket, dct["Key"])
        p._meta = {
            "Key": dct["Key"],
            "LastModified": dct["LastModified"],
            "ETag": dct["ETag"],
            "ContentLength": dct["Size"],
            "StorageClass": dct["StorageClass"],
            "Owner": dct.get("Owner", {}),
        }
        return p

    def update_metadata(self: "S3Path", metadata: dict):  # pragma: no cover
        raise NotImplementedError(
            "You CANNOT only update metadata without changing the content of the "
            "object! You can only do full replace ment via the .write_text() or "
            ".write_bytes() API. This method will NEVER be implemented!"
        )



# -*- coding: utf-8 -*-

"""
Objective Oriented Interface for AWS S3, similar to pathlib.
"""

from ._version import __version__

__short_description__ = "Objective Oriented Interface for AWS S3, similar to pathlib."
__license__ = "Apache License 2.0"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "sanhehu@amazon.com"
__github_username__ = "aws-samples"

try:
    from . import utils
    from .aws import context
    from .core import S3Path

    from iterproxy import and_, or_, not_
except ImportError:  # pragma: no cover
    pass
except:  # pragma: no cover
    raise



# -*- coding: utf-8 -*-

"""
This module implements the core OOP interface :class:`S3Path`.

Import::

    >>> from s3pathlib.core import S3Path
    # or
    >>> from s3pathlib import S3Path
"""

from .s3path import S3Path



# -*- coding: utf-8 -*-

"""
Many popular ORM framework offers a popular feature that can use the following
syntax to filter an iterable of object by its attributes. For example, in
`sqlalchemy <https://docs.sqlalchemy.org/en/14/orm/quickstart.html#simple-select>`_
you can use:

.. code-block:: python

    class User(Base):
        name: str

    select(User).where(User.name == "alice"))

This module implements the same feature.
"""

import typing as T
import functools

FilterableType = T.TypeVar("FilterableType")


class FilterableProperty(T.Generic[FilterableType]):
    """
    A descriptor decorator that convert a method to a property method.
    ALSO, convert the class attribute to be a comparable object that returns
    filter function for :class:`~s3pathlib.iterproxy.IterProxy` to use.

    .. code-block:: python

        class User:
            def __init__(self, name: str):
                self.name = name

            @FilterableProperty
            def username(self) -> str:
                return self.name

        filter_function = User.username == "alice
        assert filter_function(User(name="alice")) == True
        assert filter_function(User(name="bob")) == False
    """

    def __init__(self, func: callable):
        functools.wraps(func)(self)
        self._func = func

    def __get__(self, obj: T.Union['FilterableType', None], obj_type):
        if obj is None:
            return self
        return self._func(obj)

    def __eq__(self, other):
        def filter_(obj):
            return self._func(obj) == other

        return filter_

    def __ne__(self, other):
        def filter_(obj):
            return self._func(obj) != other

        return filter_

    def __gt__(self, other):
        def filter_(obj):
            return self._func(obj) > other

        return filter_

    def __lt__(self, other):
        def filter_(obj):
            return self._func(obj) < other

        return filter_

    def __ge__(self, other):
        def filter_(obj):
            return self._func(obj) >= other

        return filter_

    def __le__(self, other):
        def filter_(obj):
            return self._func(obj) <= other

        return filter_

    def equal_to(self, other):  # pragma: no cover
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name == ``other``

        .. versionadded:: 1.0.3
        """
        return self.__eq__(other)

    def not_equal_to(self, other):  # pragma: no cover
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name != ``other``

        .. versionadded:: 1.0.4
        """
        return self.__ne__(other)

    def greater(self, other):  # pragma: no cover
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name > ``other``

        .. versionadded:: 1.0.4
        """
        return self.__gt__(other)

    def less(self, other):  # pragma: no cover
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name < ``other``

        .. versionadded:: 1.0.4
        """
        return self.__lt__(other)

    def greater_equal(self, other):  # pragma: no cover
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name >= ``other``

        .. versionadded:: 1.0.4
        """
        return self.__eq__(other)

    def less_equal(self, other):  # pragma: no cover
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name <= ``other``

        .. versionadded:: 1.0.4
        """
        return self.__eq__(other)

    def between(self, lower, upper):
        """
        Return a filter function that returns True
        only if ``lower <= FilterableType.attribute_name <= upper``

        .. versionadded:: 1.0.3
        """

        def filter_(obj):
            return lower <= self._func(obj) <= upper

        return filter_

    def startswith(self, other: str):
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name.startswith(other)``.
        The attribute has to be a string attribute.

        .. versionadded:: 1.0.3
        """

        def filter_(obj):
            return self._func(obj).startswith(other)

        return filter_

    def endswith(self, other: str):
        """
        Return a filter function that returns True
        only if ``FilterableType.attribute_name.endswith(other)``.
        The attribute has to be a string attribute.

        .. versionadded:: 1.0.3
        """

        def filter_(obj):
            return self._func(obj).endswith(other)

        return filter_

    def contains(self, other):
        """
        Return a filter function that returns True
        only if ``other in FilterableType.attribute_name``

        .. versionadded:: 1.0.3
        """

        def filter_(obj):
            return other in self._func(obj)

        return filter_



# -*- coding: utf-8 -*-

"""
Copy file from s3 to s3.
"""

import typing as T
from datetime import datetime

from func_args import NOTHING

from .. import client as better_client
from ..type import TagType, MetadataType

from .resolve_s3_client import resolve_s3_client
from ..aws import context

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class CopyAPIMixin:
    """
    A mixin class that implements copy method.
    """

    def copy_file(
        self: "S3Path",
        dst: "S3Path",
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        overwrite: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        copy_source_if_match: str = NOTHING,
        copy_source_if_modified_since: datetime = NOTHING,
        copy_source_if_none_match: str = NOTHING,
        copy_source_if_unmodified_since: datetime = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        copy_source_sse_customer_algorithm: str = NOTHING,
        copy_source_sse_customer_key: str = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
        expected_source_bucket_owner: str = NOTHING,
    ) -> dict:
        """
        Copy an S3 directory to a different S3 directory, including all
        sub-directory and files.

        :param dst: copy to s3 object, it has to be an object
        :param overwrite: if False, none of the file will be uploaded / overwritten
            if any of target s3 location already taken.

        :return: number of object are copied, 0 or 1.

        .. versionadded:: 1.0.1

        .. versionchanged:: 1.3.1

            add ``metadata`` and ``tags`` argument
        """
        self.ensure_object()
        dst.ensure_object()
        self.ensure_not_relpath()
        dst.ensure_not_relpath()
        if overwrite is False:
            dst.ensure_not_exists(bsm=bsm)
        s3_client = resolve_s3_client(context, bsm)
        return better_client.copy_object(
            s3_client,
            src_bucket=self.bucket,
            src_key=self.key,
            dst_bucket=dst.bucket,
            dst_key=dst.key,
            metadata=metadata,
            tags=tags,
            acl=acl,
            cache_control=cache_control,
            content_disposition=content_disposition,
            content_encoding=content_encoding,
            content_language=content_language,
            content_md5=content_md5,
            content_type=content_type,
            copy_source_if_match=copy_source_if_match,
            copy_source_if_modified_since=copy_source_if_modified_since,
            copy_source_if_none_match=copy_source_if_none_match,
            copy_source_if_unmodified_since=copy_source_if_unmodified_since,
            expires_datetime=expires_datetime,
            grant_full_control=grant_full_control,
            grant_read=grant_read,
            grant_read_acp=grant_read_acp,
            grant_write_acp=grant_write_acp,
            server_side_encryption=server_side_encryption,
            storage_class=storage_class,
            website_redirect_location=website_redirect_location,
            sse_customer_algorithm=sse_customer_algorithm,
            sse_customer_key=sse_customer_key,
            sse_kms_key_id=sse_kms_key_id,
            sse_kms_encryption_context=sse_kms_encryption_context,
            bucket_key_enabled=bucket_key_enabled,
            copy_source_sse_customer_algorithm=copy_source_sse_customer_algorithm,
            copy_source_sse_customer_key=copy_source_sse_customer_key,
            request_payer=request_payer,
            object_lock_mode=object_lock_mode,
            object_lock_retain_until_datetime=object_lock_retain_until_datetime,
            object_lock_legal_hold_status=object_lock_legal_hold_status,
            expected_bucket_owner=expected_bucket_owner,
            expected_source_bucket_owner=expected_source_bucket_owner,
        )

    def copy_dir(
        self: "S3Path",
        dst: "S3Path",
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        overwrite: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        copy_source_if_match: str = NOTHING,
        copy_source_if_modified_since: datetime = NOTHING,
        copy_source_if_none_match: str = NOTHING,
        copy_source_if_unmodified_since: datetime = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        copy_source_sse_customer_algorithm: str = NOTHING,
        copy_source_sse_customer_key: str = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
        expected_source_bucket_owner: str = NOTHING,
    ):
        """
        Copy an S3 directory to a different S3 directory, including all
        sub-directory and files.

        :param dst: copy to s3 directory, it has to be a directory
        :param overwrite: if False, none of the file will be uploaded / overwritten
            if any of target s3 location already taken.

        :return: number of objects are copied

        .. versionadded:: 1.0.1
        """
        self.ensure_dir()
        dst.ensure_dir()
        self.ensure_not_relpath()
        dst.ensure_not_relpath()
        todo: T.List[T.Tuple["S3Path", "S3Path"]] = list()
        for p_src in self.iter_objects(bsm=bsm):
            p_relpath = p_src.relative_to(self)
            p_dst = dst.joinpath(p_relpath)
            todo.append((p_src, p_dst))

        if overwrite is False:
            for p_src, p_dst in todo:
                p_dst.ensure_not_exists(bsm=bsm)

        for p_src, p_dst in todo:
            p_src.copy_file(
                p_dst,
                metadata=metadata,
                tags=tags,
                overwrite=True,
                bsm=bsm,
                acl=acl,
                cache_control=cache_control,
                content_disposition=content_disposition,
                content_encoding=content_encoding,
                content_language=content_language,
                content_md5=content_md5,
                content_type=content_type,
                copy_source_if_match=copy_source_if_match,
                copy_source_if_modified_since=copy_source_if_modified_since,
                copy_source_if_none_match=copy_source_if_none_match,
                copy_source_if_unmodified_since=copy_source_if_unmodified_since,
                expires_datetime=expires_datetime,
                grant_full_control=grant_full_control,
                grant_read=grant_read,
                grant_read_acp=grant_read_acp,
                grant_write_acp=grant_write_acp,
                server_side_encryption=server_side_encryption,
                storage_class=storage_class,
                website_redirect_location=website_redirect_location,
                sse_customer_algorithm=sse_customer_algorithm,
                sse_customer_key=sse_customer_key,
                sse_kms_key_id=sse_kms_key_id,
                sse_kms_encryption_context=sse_kms_encryption_context,
                bucket_key_enabled=bucket_key_enabled,
                copy_source_sse_customer_algorithm=copy_source_sse_customer_algorithm,
                copy_source_sse_customer_key=copy_source_sse_customer_key,
                request_payer=request_payer,
                object_lock_mode=object_lock_mode,
                object_lock_retain_until_datetime=object_lock_retain_until_datetime,
                object_lock_legal_hold_status=object_lock_legal_hold_status,
                expected_bucket_owner=expected_bucket_owner,
                expected_source_bucket_owner=expected_source_bucket_owner,
            )

        return len(todo)

    def copy_to(
        self: "S3Path",
        dst: "S3Path",
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        overwrite: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        copy_source_if_match: str = NOTHING,
        copy_source_if_modified_since: datetime = NOTHING,
        copy_source_if_none_match: str = NOTHING,
        copy_source_if_unmodified_since: datetime = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        copy_source_sse_customer_algorithm: str = NOTHING,
        copy_source_sse_customer_key: str = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
        expected_source_bucket_owner: str = NOTHING,
    ) -> int:
        """
        Copy s3 object or s3 directory from one place to another place.

        :param dst: copy to s3 path
        :param overwrite: if False, non of the file will be upload / overwritten
            if any of target s3 location already taken.

        .. versionadded:: 1.0.1

        .. versionchanged:: 1.3.1

            add ``metadata`` and ``tags`` argument
        """
        if self.is_dir():
            return self.copy_dir(
                dst=dst,
                metadata=metadata,
                tags=tags,
                overwrite=overwrite,
                bsm=bsm,
                acl=acl,
                cache_control=cache_control,
                content_disposition=content_disposition,
                content_encoding=content_encoding,
                content_language=content_language,
                content_md5=content_md5,
                content_type=content_type,
                copy_source_if_match=copy_source_if_match,
                copy_source_if_modified_since=copy_source_if_modified_since,
                copy_source_if_none_match=copy_source_if_none_match,
                copy_source_if_unmodified_since=copy_source_if_unmodified_since,
                expires_datetime=expires_datetime,
                grant_full_control=grant_full_control,
                grant_read=grant_read,
                grant_read_acp=grant_read_acp,
                grant_write_acp=grant_write_acp,
                server_side_encryption=server_side_encryption,
                storage_class=storage_class,
                website_redirect_location=website_redirect_location,
                sse_customer_algorithm=sse_customer_algorithm,
                sse_customer_key=sse_customer_key,
                sse_kms_key_id=sse_kms_key_id,
                sse_kms_encryption_context=sse_kms_encryption_context,
                bucket_key_enabled=bucket_key_enabled,
                copy_source_sse_customer_algorithm=copy_source_sse_customer_algorithm,
                copy_source_sse_customer_key=copy_source_sse_customer_key,
                request_payer=request_payer,
                object_lock_mode=object_lock_mode,
                object_lock_retain_until_datetime=object_lock_retain_until_datetime,
                object_lock_legal_hold_status=object_lock_legal_hold_status,
                expected_bucket_owner=expected_bucket_owner,
                expected_source_bucket_owner=expected_source_bucket_owner,
            )
        elif self.is_file():
            self.copy_file(
                dst=dst,
                overwrite=overwrite,
                bsm=bsm,
                metadata=metadata,
                tags=tags,
                acl=acl,
                cache_control=cache_control,
                content_disposition=content_disposition,
                content_encoding=content_encoding,
                content_language=content_language,
                content_md5=content_md5,
                content_type=content_type,
                copy_source_if_match=copy_source_if_match,
                copy_source_if_modified_since=copy_source_if_modified_since,
                copy_source_if_none_match=copy_source_if_none_match,
                copy_source_if_unmodified_since=copy_source_if_unmodified_since,
                expires_datetime=expires_datetime,
                grant_full_control=grant_full_control,
                grant_read=grant_read,
                grant_read_acp=grant_read_acp,
                grant_write_acp=grant_write_acp,
                server_side_encryption=server_side_encryption,
                storage_class=storage_class,
                website_redirect_location=website_redirect_location,
                sse_customer_algorithm=sse_customer_algorithm,
                sse_customer_key=sse_customer_key,
                sse_kms_key_id=sse_kms_key_id,
                sse_kms_encryption_context=sse_kms_encryption_context,
                bucket_key_enabled=bucket_key_enabled,
                copy_source_sse_customer_algorithm=copy_source_sse_customer_algorithm,
                copy_source_sse_customer_key=copy_source_sse_customer_key,
                request_payer=request_payer,
                object_lock_mode=object_lock_mode,
                object_lock_retain_until_datetime=object_lock_retain_until_datetime,
                object_lock_legal_hold_status=object_lock_legal_hold_status,
                expected_bucket_owner=expected_bucket_owner,
                expected_source_bucket_owner=expected_source_bucket_owner,
            )
            return 1
        else:  # pragma: no cover
            raise TypeError

    def move_to(
        self: "S3Path",
        dst: "S3Path",
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        overwrite: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        copy_source_if_match: str = NOTHING,
        copy_source_if_modified_since: datetime = NOTHING,
        copy_source_if_none_match: str = NOTHING,
        copy_source_if_unmodified_since: datetime = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        copy_source_sse_customer_algorithm: str = NOTHING,
        copy_source_sse_customer_key: str = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
        expected_source_bucket_owner: str = NOTHING,
    ) -> int:
        """
        Move s3 object or s3 directory from one place to another place. It is
        firstly :meth:`S3Path.copy_to` then :meth:`S3Path.delete_if_exists`

        :param dst: copy to s3 path
        :param overwrite: if False, non of the file will be upload / overwritten
            if any of target s3 location already taken.

        .. versionadded:: 1.0.1

        .. versionchanged:: 1.3.1

            add ``metadata`` and ``tags`` argument
        """
        count = self.copy_to(
            dst=dst,
            metadata=metadata,
            tags=tags,
            overwrite=overwrite,
            bsm=bsm,
            acl=acl,
            cache_control=cache_control,
            content_disposition=content_disposition,
            content_encoding=content_encoding,
            content_language=content_language,
            content_md5=content_md5,
            content_type=content_type,
            copy_source_if_match=copy_source_if_match,
            copy_source_if_modified_since=copy_source_if_modified_since,
            copy_source_if_none_match=copy_source_if_none_match,
            copy_source_if_unmodified_since=copy_source_if_unmodified_since,
            expires_datetime=expires_datetime,
            grant_full_control=grant_full_control,
            grant_read=grant_read,
            grant_read_acp=grant_read_acp,
            grant_write_acp=grant_write_acp,
            server_side_encryption=server_side_encryption,
            storage_class=storage_class,
            website_redirect_location=website_redirect_location,
            sse_customer_algorithm=sse_customer_algorithm,
            sse_customer_key=sse_customer_key,
            sse_kms_key_id=sse_kms_key_id,
            sse_kms_encryption_context=sse_kms_encryption_context,
            bucket_key_enabled=bucket_key_enabled,
            copy_source_sse_customer_algorithm=copy_source_sse_customer_algorithm,
            copy_source_sse_customer_key=copy_source_sse_customer_key,
            request_payer=request_payer,
            object_lock_mode=object_lock_mode,
            object_lock_retain_until_datetime=object_lock_retain_until_datetime,
            object_lock_legal_hold_status=object_lock_legal_hold_status,
            expected_bucket_owner=expected_bucket_owner,
            expected_source_bucket_owner=expected_source_bucket_owner,
        )
        self.delete_if_exists(bsm=bsm)
        return count



# -*- coding: utf-8 -*-

"""
Is the S3Path a XYZ testing.

- :meth:`~IsTestAPIMixin.is_void`
- :meth:`~IsTestAPIMixin.is_dir`
- :meth:`~IsTestAPIMixin.is_file`
- :meth:`~IsTestAPIMixin.is_bucket`
"""

import typing as T

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class IsTestAPIMixin:
    """
    A mixin class that implements the condition test methods.
    """

    def is_void(self: "S3Path") -> bool:
        """
        Test if it is a void S3 path.

        **Definition**

        no bucket, no key, no parts, no type, no nothing.
        A void path is also a special :meth:`relative path <is_relpath>`,
        because any path join with void path results to itself.
        """
        return (self._bucket is None) and (len(self._parts) == 0)

    def is_dir(self: "S3Path") -> bool:
        """
        Test if it is a S3 directory

        **Definition**

        A logical S3 directory that never physically exists. An S3
        :meth:`bucket <is_bucket>` is also a special dir, which is the root dir.

        .. versionadded:: 1.0.1
        """
        if self._is_dir is None:
            return False
        else:
            return self._is_dir

    def is_file(self: "S3Path") -> bool:
        """
        Test if it is a S3 object

        **Definition**

        A S3 object.

        .. versionadded:: 1.0.1
        """
        if self._is_dir is None:
            return False
        else:
            return not self._is_dir

    def is_bucket(self: "S3Path") -> bool:
        """
        Test if it is a S3 bucket.

        **Definition**

        A S3 bucket, the root folder S3 path is equivalent to a S3 bucket.
        A S3 bucket are always :meth:`is_dir() is True <is_dir>`

        .. versionadded:: 1.0.1
        """
        return (
            (self._bucket is not None)
            and (len(self._parts) == 0)
            and (self._is_dir is True)
        )

    def ensure_object(self: "S3Path") -> None:
        """
        A validator method that ensure it represents a S3 object.

        .. versionadded:: 1.0.1
        """
        if self.is_file() is not True:
            raise TypeError(f"S3 URI: {self} IS NOT a valid s3 object!")

    def ensure_file(self: "S3Path") -> None:
        """
        A validator method that ensure it represents a S3 object.

        .. versionadded:: 1.2.1
        """
        return self.ensure_object()

    def ensure_not_object(self: "S3Path") -> None:
        """
        A validator method that ensure it IS NOT a S3 object.

        .. versionadded:: 1.2.1
        """
        if self.is_file() is True:
            raise TypeError(f"S3 URI: {self} IS an s3 object!")

    def ensure_not_file(self: "S3Path") -> None:
        """
        A validator method that ensure it IS NOT a S3 object.

        .. versionadded:: 1.2.1
        """
        self.ensure_not_object()

    def ensure_dir(self: "S3Path") -> None:
        """
        A validator method that ensure it represents a S3 dir.

        .. versionadded:: 1.0.1
        """
        if self.is_dir() is not True:
            raise TypeError(f"{self} IS NOT a valid s3 directory!")

    def ensure_not_dir(self: "S3Path") -> None:
        """
        A validator method that ensure it IS NOT a S3 dir.

        .. versionadded:: 1.2.1
        """
        if self.is_dir() is True:
            raise TypeError(f"{self} IS a s3 directory!")



# -*- coding: utf-8 -*-

"""
Manage the AWS environment that s3pathlib dealing with.
"""

from typing import Optional

try:
    import boto3
except ImportError:  # pragma: no cover
    pass
except:  # pragma: no cover
    raise


class Context:
    """
    A globally available context object managing AWS SDK credentials.

    TODO: use singleton pattern to create context object
    """

    def __init__(self):
        self.boto_ses: Optional['boto3.session.Session'] = None
        self._aws_region: Optional[str] = None
        self._aws_account_id: Optional[str] = None
        self._s3_client = None
        self._sts_client = None

        # try to create default session
        try:
            self.boto_ses = boto3.session.Session()
        except:  # pragma: no cover
            pass

    def attach_boto_session(self, boto_ses):
        """
        Attach a custom boto session.

        :type boto_ses: boto3.session.Session
        """
        self.boto_ses = boto_ses
        self._s3_client = None
        self._sts_client = None

    @property
    def s3_client(self):
        """
        Access the s3 client.

        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#client
        """
        if self._s3_client is None:
            self._s3_client = self.boto_ses.client("s3")
        return self._s3_client

    @property
    def sts_client(self):
        """
        Access the s3 client.

        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#client
        """
        if self._sts_client is None:
            self._sts_client = self.boto_ses.client("sts")
        return self._sts_client

    @property
    def aws_account_id(self) -> str:
        """
        The AWS Account ID of the current boto session/
        """
        if self._aws_account_id is None:
            self._aws_account_id = self.sts_client.get_caller_identity()["Account"]
        return self._aws_account_id

    @property
    def aws_region(self) -> str:
        if self._aws_region is None:
            self._aws_region = self.boto_ses.region_name
        return self._aws_region


context = Context()



# -*- coding: utf-8 -*-

"""
Tagging related API.
"""

import typing as T

from .resolve_s3_client import resolve_s3_client
from .. import client as better_client
from ..aws import context
from ..type import TagType

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class TaggingAPIMixin:
    """
    A mixin class that implements the tagging related methods.
    """

    def get_tags(
        self: "S3Path",
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> TagType:
        """
        Get s3 object tags in key value pairs dict.

        :return: the s3 object tags in string key value pairs dict.

        .. versionadded:: 1.1.1
        """
        self.ensure_object()
        s3_client = resolve_s3_client(context, bsm)
        return better_client.get_object_tagging(s3_client, self.bucket, self.key)

    def put_tags(
        self: "S3Path",
        tags: TagType,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> TagType:
        """
        Do full replacement of s3 object tags.

        :param tags: the s3 object tags in string key value pairs dict.

        :return: the s3 object tags in string key value pairs dict.

        .. versionadded:: 1.1.1
        """
        self.ensure_object()
        s3_client = resolve_s3_client(context, bsm)
        better_client.put_object_tagging(s3_client, self.bucket, self.key, tags)
        return tags

    def update_tags(
        self: "S3Path",
        tags: TagType,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> TagType:
        """
        Do partial updates of s3 object tags.

        :param tags: the s3 object tags in string key value pairs dict.

        :return: the latest, merged object tags in string key value pairs dict.

        .. versionadded:: 1.1.1
        """
        self.ensure_object()
        s3_client = resolve_s3_client(context, bsm)
        return better_client.update_object_tagging(
            s3_client,
            self.bucket,
            self.key,
            tags,
        )



# -*- coding: utf-8 -*-

"""
Tagging related API.
"""

import typing as T

from iterproxy import IterProxy

from .resolve_s3_client import resolve_s3_client
from .. import utils
from ..aws import context

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class S3PathIterProxy(IterProxy):
    """
    An iterator proxy utility class provide client side in-memory filter.
    It is highly inspired by sqlalchemy Result Proxy that depends on SQL server
    side filter.

    Allow client side in-memory filtering for iterator object that yield
    :class:`S3Path`.

    It is a special variation of :class:`s3pathlib.iterproxy.IterProxy`,
    See :class:`s3pathlib.iterproxy.IterProxy` for more details

    .. versionadded:: 1.0.3
    """

    def __next__(self) -> "S3Path":
        return super(S3PathIterProxy, self).__next__()

    def one(self) -> "S3Path":
        return super(S3PathIterProxy, self).one()

    def one_or_none(self) -> T.Union["S3Path", None]:
        return super(S3PathIterProxy, self).one_or_none()

    def many(self, k: int) -> T.List["S3Path"]:
        return super(S3PathIterProxy, self).many(k)

    def all(self) -> T.List["S3Path"]:
        return super(S3PathIterProxy, self).all()

    def skip(self, k: int) -> "S3PathIterProxy":
        return super(S3PathIterProxy, self).skip(k=k)

    def filter_by_ext(self, *exts: str) -> "S3PathIterProxy":
        """
        Filter S3 object by file extension. Case is insensitive.

        Example::

            >>> p = S3Path("bucket")
            >>> for path in p.iter_objects().filter_by_ext(".csv", ".json"):
            ...      print(path)
        """
        n = len(exts)
        if n == 0:
            raise ValueError
        elif n == 1:
            ext = exts[0].lower()

            def f(p: 'S3Path') -> bool:
                return p.ext.lower() == ext

            return self.filter(f)
        else:
            valid_exts = set([ext.lower() for ext in exts])

            def f(p: 'S3Path') -> bool:
                return p.ext.lower() in valid_exts

            return self.filter(f)


class IterObjectsAPIMixin:
    """
    A mixin class that implements the iter objects methods.
    """

    def _iter_objects(
        self: "S3Path",
        batch_size: int = 1000,
        limit: int = None,
        recursive: bool = True,
        include_folder: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> T.Iterable["S3Path"]:
        s3_client = resolve_s3_client(context, bsm)
        for dct in utils.iter_objects(
            s3_client=s3_client,
            bucket=self.bucket,
            prefix=self.key,
            batch_size=batch_size,
            limit=limit,
            recursive=recursive,
            include_folder=include_folder,
        ):
            yield self._from_content_dict(
                bucket=self.bucket,
                dct=dct,
            )

    def iter_objects(
        self: "S3Path",
        batch_size: int = 1000,
        limit: int = None,
        recursive: bool = True,
        include_folder: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> S3PathIterProxy:
        """
        Recursively iterate objects under this prefix, yield :class:`S3Path`.

        :param batch_size: number of s3 object returned per paginator,
            valid value is from 1 ~ 1000. large number can reduce IO.
        :param limit: total number of s3 object to return
        :param recursive: if True, it won't include files in sub folders
        :param include_folder: AWS S3 consider object that key endswith "/"
            and size = 0 as a logical folder. But physically it is still object.
            By default ``list_objects_v2`` API returns logical folder object,
            you can use this flag to filter it out.

        .. versionadded:: 1.0.1

        TODO: add unix glob liked syntax for pattern matching
        """
        return S3PathIterProxy(
            iterable=self._iter_objects(
                batch_size=batch_size,
                limit=limit,
                recursive=recursive,
                include_folder=include_folder,
                bsm=bsm,
            )
        )

    def _iterdir(
        self: "S3Path",
        batch_size: int = 1000,
        limit: int = None,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> T.Iterable["S3Path"]:
        s3_client = resolve_s3_client(context, bsm)
        paginator = s3_client.get_paginator("list_objects_v2")
        pagination_config = dict(PageSize=batch_size)
        if limit:  # pragma: no cover
            pagination_config["MaxItems"] = limit

        for res in paginator.paginate(
            Bucket=self.bucket,
            Prefix=self.key,
            Delimiter="/",
            PaginationConfig=pagination_config,
        ):
            for dct in res.get("CommonPrefixes", list()):
                yield self.root.joinpath(dct["Prefix"])

            for dct in res.get("Contents", list()):
                yield self._from_content_dict(self.bucket, dct)

    def iterdir(
        self: "S3Path",
        batch_size: int = 1000,
        limit: int = None,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> S3PathIterProxy:
        """
        iterate objects and folder under this prefix non-recursively,
        yield :class:`S3Path`.

        :param batch_size: number of s3 object returned per paginator,
            valid value is from 1 ~ 1000. large number can reduce IO.
        :param limit: total number of s3 object (not folder)to return

        .. versionadded:: 1.0.6
        """
        return S3PathIterProxy(
            iterable=self._iterdir(
                batch_size=batch_size,
                limit=limit,
                bsm=bsm,
            )
        )

    def calculate_total_size(
        self: "S3Path",
        for_human: bool = False,
        include_folder: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> T.Tuple[int, T.Union[int, str]]:
        """
        Perform the "Calculate Total Size" action in AWS S3 console

        :param for_human: if true, returns human readable string for "size"
        :param include_folder: see :meth:`iter_objects`

        :return: a tuple, first value is number of objects,
            second value is total size in bytes

        .. versionadded:: 1.0.1
        """
        self.ensure_dir()
        s3_client = resolve_s3_client(context, bsm)
        count, size = utils.calculate_total_size(
            s3_client=s3_client,
            bucket=self.bucket,
            prefix=self.key,
            include_folder=include_folder,
        )
        if for_human:
            size = utils.repr_data_size(size)
        return count, size

    def count_objects(
        self: "S3Path",
        include_folder: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> int:
        """
        Count how many objects are under this s3 directory.

        :param include_folder: see :meth:`iter_objects`

        :return: an integer represents the number of objects

        .. versionadded:: 1.0.1
        """
        self.ensure_dir()
        s3_client = resolve_s3_client(context, bsm)
        return utils.count_objects(
            s3_client=s3_client,
            bucket=self.bucket,
            prefix=self.key,
            include_folder=include_folder,
        )



# -*- coding: utf-8 -*-

"""
Upload file from local to s3.
"""

import typing as T

from pathlib_mate import Path

from .resolve_s3_client import resolve_s3_client
from .. import utils
from ..type import PathType
from ..aws import context

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class UploadAPIMixin:
    """
    A mixin class that implements upload method.
    """

    def upload_file(
        self: "S3Path",
        path: PathType,
        overwrite: bool = False,
        extra_args: dict = None,
        callback: callable = None,
        config=None,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> dict:
        """
        Upload a file from local file system to targeted S3 path

        Example::

            >>> s3path = S3Path("bucket", "artifacts", "deployment.zip")
            >>> s3path.upload_file(path="/tmp/build/deployment.zip", overwrite=True)

        :param path: absolute path of the file on the local file system
            you want to upload
        :param overwrite: if False, non of the file will be upload / overwritten
            if any of target s3 location already taken.

        .. versionadded:: 1.0.1
        """
        self.ensure_object()
        if overwrite is False:
            self.ensure_not_exists(bsm=bsm)
        p = Path(path)
        s3_client = resolve_s3_client(context, bsm)
        return s3_client.upload_file(
            p.abspath,
            Bucket=self.bucket,
            Key=self.key,
            ExtraArgs=extra_args,
            Callback=callback,
            Config=config,
        )

    def upload_dir(
        self: "S3Path",
        local_dir: PathType,
        pattern: str = "**/*",
        overwrite: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> int:
        """
        Upload a directory on local file system and all sub-folders, files to
        a S3 prefix (logical directory)

        Example::

            >>> s3path = S3Path("bucket", "datalake", "orders/")
            >>> s3path.upload_dir(path="/data/orders", overwrite=True)

        :param local_dir: absolute path of the directory on the
            local file system you want to upload
        :param pattern: linux styled glob pattern match syntax. see this
            official reference
            https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob
            for more details
        :param overwrite: if False, non of the file will be upload / overwritten
            if any of target s3 location already taken.

        :return: number of files uploaded

        .. versionadded:: 1.0.1
        """
        self.ensure_dir()
        s3_client = resolve_s3_client(context, bsm)
        return utils.upload_dir(
            s3_client=s3_client,
            bucket=self.bucket,
            prefix=self.key,
            local_dir=local_dir,
            pattern=pattern,
            overwrite=overwrite,
        )



# -*- coding: utf-8 -*-

import typing as T
from boto_session_manager import BotoSesManager, AwsServiceEnum
from ..aws import Context


def resolve_s3_client(
    context: Context,
    bsm: T.Optional['BotoSesManager'] = None,
):
    """
    Figure out the final boto session to use for API call.
    If ``BotoSesManager`` is defined, then prioritize to use it.
    """
    if bsm is None:
        return context.s3_client
    else:
        return bsm.get_client(AwsServiceEnum.S3)



# -*- coding: utf-8 -*-

try:
    import smart_open

    smart_open_version = smart_open.__version__
    (
        smart_open_version_major,
        smart_open_version_minor,
        _,
    ) = smart_open_version.split(".")
    smart_open_version_major = int(smart_open_version_major)
    smart_open_version_minor = int(smart_open_version_minor)
except ImportError:  # pragma: no cover
    smart_open = None
    smart_open_version_major = None
    smart_open_version_minor = None
except:  # pragma: no cover
    raise


class Compat:  # pragma: no cover
    @property
    def smart_open_version_major(self) -> int:
        if smart_open_version_major is None:
            raise ImportError("You don't have smart_open installed")
        return smart_open_version_major

    @property
    def smart_open_version_minor(self) -> int:
        if smart_open_version_minor is None:
            raise ImportError("You don't have smart_open installed")
        return smart_open_version_minor


compat = Compat()



# -*- coding: utf-8 -*-

import sys
import subprocess

from .paths import dir_project_root, dir_htmlcov, bin_pytest

if sys.platform == "win32":
    open_cmd = "open"
else:
    open_cmd = "open"


def normalize_module(module: str) -> str:
    if module.endswith(".py"):
        return module[:-3]
    else:
        return module


def _run_cov_test(
    bin_pytest: str,
    script: str,
    module: str,
    root_dir: str,
    htmlcov_dir: str,
):
    """
    A simple wrapper around pytest + coverage cli command.
    :param bin_pytest: the path to pytest executable
    :param script: the path to test script
    :param module: the dot notation to the python module you want to calculate
        coverage
    :param root_dir: the dir to dump coverage results binary file
    :param htmlcov_dir: the dir to dump HTML output
    """
    module = normalize_module(module)
    args = [
        bin_pytest,
        "-s", "--tb=native",
        f"--rootdir={root_dir}",
        f"--cov={module}",
        "--cov-report", "term-missing",
        "--cov-report", f"html:{htmlcov_dir}",
        script,
    ]
    subprocess.run(args)


def run_cov_test(script: str, module: str, open_browser: bool = False):
    _run_cov_test(
        bin_pytest=f"{bin_pytest}",
        script=script,
        module=module,
        root_dir=f"{dir_project_root}",
        htmlcov_dir=f"{dir_htmlcov}",
    )
    if open_browser:
        _open_cov_report(module)


def _open_cov_report(module: str):
    module = normalize_module(module)
    parts = module.split(".")
    module_name = parts[-1]
    module_path = str(dir_project_root.joinpath(*parts)) + ".py"
    print(module_path)
    found_cov_report_html = False
    html_path = None
    for p in dir_htmlcov.glob("**/*_py.html"):
        if p.name.endswith(f"{module_name}_py.html"):
            if module_path in p.read_text():
                found_cov_report_html = True
                html_path = str(p)
                break
    if found_cov_report_html:
        subprocess.run([open_cmd, html_path])



# -*- coding: utf-8 -*-

"""
Relative Path implementation.
"""

import typing as T

from .. import utils

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class RelativePathAPIMixin:
    """
    A mixin class that implements the relative path concept.
    """

    @classmethod
    def make_relpath(
        cls: T.Type["S3Path"],
        *parts: str,
    ) -> "S3Path":
        """
        A construct method that create a relative S3 Path.

        Definition of relative path:

        - no bucket (self.bucket is None)
        - has some parts or no part. when no part, it is a special relative path.
            Any path add this relative path resulting to itself. We call this
            special relative path **Void relative path**. A
            **Void relative path** is logically equivalent to **Void s3 path**.
        - relative path can be a file (object) or a directory. The
            **Void relative path** is neither a file or a directory.

        :param parts:

        .. versionadded:: 1.0.1

        **中文文档**

        相对路径的概念是 p1 + p2 = p3, 其中 p1 和 p3 都是实际存在的路径, 而 p2 则是
        相对路径.

        相对路径的功能是如果 p3 - p1 = p2, 那么 p1 + p2 必须还能等于 p3. 有一个特殊情况是
        如果 p1 - p1 = p0, 两个相同的绝对路径之间的相对路径是 p0, 我们还是需要满足
        p1 + p0 = p1 以保证逻辑上的一致.
        """
        _parts = list()
        last_non_empty_part = None
        for part in parts:
            chunks = utils.split_parts(part)
            if len(chunks):
                last_non_empty_part = part
            for _part in chunks:
                _parts.append(_part)

        if len(_parts):
            _is_dir = last_non_empty_part.endswith("/")
        else:
            _is_dir = None

        return cls._from_parsed_parts(
            bucket=None,
            parts=_parts,
            is_dir=_is_dir,
        )

    def relative_to(self: "S3Path", other: "S3Path") -> "S3Path":
        """
        Return the relative path to another path. If the operation
        is not possible (because this is not a sub path of the other path),
        raise ``ValueError``.

        ``-`` is a syntax sugar for ``relative_to``. See more information at
        :meth:`~S3Path.__sub__`.

        The relative path usually works with :meth:`join_path` to form a new
        path. Or you can use the ``/`` syntax sugar as well. See more
        information at :meth:`~S3Path.__truediv__`.

        Examples::

            >>> S3Path("bucket", "a/b/c").relative_to(S3Path("bucket", "a")).parts
            ['b', 'c']

            >>> S3Path("bucket", "a").relative_to(S3Path("bucket", "a")).parts
            []

            >>> S3Path("bucket", "a").relative_to(S3Path("bucket", "a/b/c")).parts
            ValueError ...

        :param other: other :class:`S3Path` instance.

        :return: a relative path object, which is a special version of S3Path

        .. versionadded:: 1.0.1
        """
        if (
            (self._bucket != other._bucket)
            or (self._bucket is None)
            or (other._bucket is None)
        ):
            msg = (
                "both 'from path' {} and 'to path' {} has to be concrete path!"
            ).format(self, other)
            raise ValueError(msg)

        n = len(other._parts)
        if self._parts[:n] != other._parts:
            msg = "{} does not start with {}".format(
                self.uri,
                other.uri,
            )
            raise ValueError(msg)
        rel_parts = self._parts[n:]
        if len(rel_parts):
            is_dir = self._is_dir
        else:
            is_dir = None
        return self._from_parsed_parts(
            bucket=None,
            parts=rel_parts,
            is_dir=is_dir,
        )

    def is_relpath(self: "S3Path") -> bool:
        """
        Relative path is a special path supposed to join with other concrete
        path.

        **Definition**

        A long full path relating to its parent directory is a relative path.
        A :meth:`void <is_void>` path also a special relative path.

        .. versionadded:: 1.0.1
        """
        if self._bucket is None:
            if len(self._parts) == 0:
                if self._is_dir is None:
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False

    def __sub__(self: "S3Path", other: "S3Path") -> "S3Path":
        """
        A syntax sugar. Basically ``s3path1 - s3path2`` is equal to
        ``s3path2.relative_to(s3path1)``

        .. versionadded:: 1.0.11
        """
        return self.relative_to(other)

    def ensure_relpath(self: "S3Path") -> None:
        """
        A validator method that ensure it represents a S3 relative path.

        .. versionadded:: 1.2.1
        """
        if self.is_relpath() is not True:
            raise TypeError(f"{self} IS a s3 relative path!")

    def ensure_not_relpath(self: "S3Path") -> None:
        """
        A validator method that ensure it represents a S3 relative path.

        Can be used if you want to raise error if it is not a relative path.

        .. versionadded:: 1.0.1
        """
        if self.is_relpath() is True:
            raise TypeError(f"{self} is not a valid s3 relative path!")



# -*- coding: utf-8 -*-

"""
Read and write related API.
"""

import typing as T
from datetime import datetime

from func_args import NOTHING

from .. import utils, client as better_client
from ..type import TagType, MetadataType
from ..aws import context

from .resolve_s3_client import resolve_s3_client

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class ReadAndWriteAPIMixin:
    """
    A mixin class that implements the Text / Bytes, Read / Write methods.
    """
    def read_bytes(
        self: "S3Path",
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> bytes:
        """
        Read binary data from s3 object. A simple wrapper around
        `s3_client.get_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.get_object>`_

        .. versionchanged:: 1.1.2

            automatically store metadata in cache.
        """
        s3_client = resolve_s3_client(context, bsm)
        response = s3_client.get_object(
            Bucket=self.bucket,
            Key=self.key,
        )
        data = response["Body"].read()
        del response["Body"]
        del response["ResponseMetadata"]
        self._meta = response
        return data

    def read_text(
        self: "S3Path",
        encoding="utf-8",
        errors="strict",
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> str:
        """
        Read text data from s3 object. A simple wrapper around
        `s3_client.get_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.get_object>`_

        .. versionchanged:: 1.1.2

            automatically store metadata in cache.
        """
        data = self.read_bytes(bsm=bsm)
        return data.decode(encoding, errors=errors)

    def write_bytes(
        self: "S3Path",
        data: bytes,
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_length: int = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        checksum_algorithm: str = NOTHING,
        checksum_crc32: str = NOTHING,
        checksum_crc32c: str = NOTHING,
        checksum_sha1: str = NOTHING,
        checksum_sha256: str = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
    ):
        """
        Write binary data to s3 object. A simple wrapper around
        `s3_client.put_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object>`_

        :param data: the text you want to write.
        :param metadata: the s3 object metadata in string key value pair dict.
        :param tags: the s3 object tags in string key value pair dict.

        .. versionadded:: 1.0.3

        .. versionchanged:: 1.1.1

            allow update metadata and tags as well
        """
        s3_client = resolve_s3_client(context, bsm)
        response = better_client.put_object(
            s3_client=s3_client,
            bucket=self.bucket,
            key=self.key,
            body=data,
            metadata=metadata,
            tags=tags,
            acl=acl,
            cache_control=cache_control,
            content_disposition=content_disposition,
            content_encoding=content_encoding,
            content_language=content_language,
            content_length=content_length,
            content_md5=content_md5,
            content_type=content_type,
            checksum_algorithm=checksum_algorithm,
            checksum_crc32=checksum_crc32,
            checksum_crc32c=checksum_crc32c,
            checksum_sha1=checksum_sha1,
            checksum_sha256=checksum_sha256,
            expires_datetime=expires_datetime,
            grant_full_control=grant_full_control,
            grant_read=grant_read,
            grant_read_acp=grant_read_acp,
            grant_write_acp=grant_write_acp,
            server_side_encryption=server_side_encryption,
            storage_class=storage_class,
            website_redirect_location=website_redirect_location,
            sse_customer_algorithm=sse_customer_algorithm,
            sse_customer_key=sse_customer_key,
            sse_kms_key_id=sse_kms_key_id,
            sse_kms_encryption_context=sse_kms_encryption_context,
            bucket_key_enabled=bucket_key_enabled,
            request_payer=request_payer,
            object_lock_mode=object_lock_mode,
            object_lock_retain_until_datetime=object_lock_retain_until_datetime,
            object_lock_legal_hold_status=object_lock_legal_hold_status,
            expected_bucket_owner=expected_bucket_owner,
        )
        self._meta = response
        self._meta["Metadata"] = metadata
        return response

    def write_text(
        self: "S3Path",
        data: str,
        encoding="utf-8",
        errors="strict",
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_length: int = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        checksum_algorithm: str = NOTHING,
        checksum_crc32: str = NOTHING,
        checksum_crc32c: str = NOTHING,
        checksum_sha1: str = NOTHING,
        checksum_sha256: str = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
    ):
        """
        Write text to s3 object. A simple wrapper around
        `s3_client.put_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object>`_

        :param data: the text you want to write.
        :param encoding: how do you want to encode text?
        :param errors: how do you want to handle encode error? can be 'strict',
            ``ignore``, ``replace``, ``xmlcharrefreplace``, ``backslashreplace``.
            see more details `here <https://docs.python.org/3/library/stdtypes.html#str.encode>`_.
        :param metadata: the s3 object metadata in string key value pair dict.
        :param tags: the s3 object tags in string key value pair dict.

        .. versionadded:: 1.0.3

        .. versionchanged:: 1.1.1

            allow update metadata and tags as well
        """
        body = data.encode(encoding, errors=errors)
        return self.write_bytes(
            data=body,
            metadata=metadata,
            tags=tags,
            bsm=bsm,
            acl=acl,
            cache_control=cache_control,
            content_disposition=content_disposition,
            content_encoding=content_encoding,
            content_language=content_language,
            content_length=content_length,
            content_md5=content_md5,
            content_type=content_type,
            checksum_algorithm=checksum_algorithm,
            checksum_crc32=checksum_crc32,
            checksum_crc32c=checksum_crc32c,
            checksum_sha1=checksum_sha1,
            checksum_sha256=checksum_sha256,
            expires_datetime=expires_datetime,
            grant_full_control=grant_full_control,
            grant_read=grant_read,
            grant_read_acp=grant_read_acp,
            grant_write_acp=grant_write_acp,
            server_side_encryption=server_side_encryption,
            storage_class=storage_class,
            website_redirect_location=website_redirect_location,
            sse_customer_algorithm=sse_customer_algorithm,
            sse_customer_key=sse_customer_key,
            sse_kms_key_id=sse_kms_key_id,
            sse_kms_encryption_context=sse_kms_encryption_context,
            bucket_key_enabled=bucket_key_enabled,
            request_payer=request_payer,
            object_lock_mode=object_lock_mode,
            object_lock_retain_until_datetime=object_lock_retain_until_datetime,
            object_lock_legal_hold_status=object_lock_legal_hold_status,
            expected_bucket_owner=expected_bucket_owner,
        )

    def touch(
        self: "S3Path",
        exist_ok: bool = True,
        metadata: T.Optional[MetadataType] = NOTHING,
        tags: T.Optional[TagType] = NOTHING,
        bsm: T.Optional["BotoSesManager"] = None,
        acl: str = NOTHING,
        cache_control: str = NOTHING,
        content_disposition: str = NOTHING,
        content_encoding: str = NOTHING,
        content_language: str = NOTHING,
        content_length: int = NOTHING,
        content_md5: str = NOTHING,
        content_type: str = NOTHING,
        checksum_algorithm: str = NOTHING,
        checksum_crc32: str = NOTHING,
        checksum_crc32c: str = NOTHING,
        checksum_sha1: str = NOTHING,
        checksum_sha256: str = NOTHING,
        expires_datetime: datetime = NOTHING,
        grant_full_control: str = NOTHING,
        grant_read: str = NOTHING,
        grant_read_acp: str = NOTHING,
        grant_write_acp: str = NOTHING,
        server_side_encryption: str = NOTHING,
        storage_class: str = NOTHING,
        website_redirect_location: str = NOTHING,
        sse_customer_algorithm: str = NOTHING,
        sse_customer_key: str = NOTHING,
        sse_kms_key_id: str = NOTHING,
        sse_kms_encryption_context: str = NOTHING,
        bucket_key_enabled: bool = NOTHING,
        request_payer: str = NOTHING,
        object_lock_mode: str = NOTHING,
        object_lock_retain_until_datetime: datetime = NOTHING,
        object_lock_legal_hold_status: str = NOTHING,
        expected_bucket_owner: str = NOTHING,
    ):
        """
        Create an empty S3 object at the S3 location if the S3 object not exists.
        Do nothing if already exists.

        :param exist_ok: if True, it won't raise error when the S3 object
            already exists.

        .. versionchanged:: 1.0.6

        .. versionchanged:: 1.2.1
        """
        self.ensure_object()

        if self.exists(bsm=bsm):
            if exist_ok:
                pass
            else:
                raise FileExistsError
        else:
            self.write_text(
                "",
                metadata=metadata,
                tags=tags,
                bsm=bsm,
                acl=acl,
                cache_control=cache_control,
                content_disposition=content_disposition,
                content_encoding=content_encoding,
                content_language=content_language,
                content_length=content_length,
                content_md5=content_md5,
                content_type=content_type,
                checksum_algorithm=checksum_algorithm,
                checksum_crc32=checksum_crc32,
                checksum_crc32c=checksum_crc32c,
                checksum_sha1=checksum_sha1,
                checksum_sha256=checksum_sha256,
                expires_datetime=expires_datetime,
                grant_full_control=grant_full_control,
                grant_read=grant_read,
                grant_read_acp=grant_read_acp,
                grant_write_acp=grant_write_acp,
                server_side_encryption=server_side_encryption,
                storage_class=storage_class,
                website_redirect_location=website_redirect_location,
                sse_customer_algorithm=sse_customer_algorithm,
                sse_customer_key=sse_customer_key,
                sse_kms_key_id=sse_kms_key_id,
                sse_kms_encryption_context=sse_kms_encryption_context,
                bucket_key_enabled=bucket_key_enabled,
                request_payer=request_payer,
                object_lock_mode=object_lock_mode,
                object_lock_retain_until_datetime=object_lock_retain_until_datetime,
                object_lock_legal_hold_status=object_lock_legal_hold_status,
                expected_bucket_owner=expected_bucket_owner,
            )

    def mkdir(
        self: "S3Path",
        exist_ok: bool = False,
        parents: bool = False,
        bsm: T.Optional["BotoSesManager"] = None,
    ):
        """
        Make an S3 folder (empty "/" file)

        .. versionchanged:: 1.0.6
        """
        if not self.is_dir():
            raise ValueError

        s3_client = resolve_s3_client(context, bsm)
        dct = utils.head_object_if_exists(
            s3_client=s3_client,
            bucket=self.bucket,
            key=self.key,
        )
        if dct:
            if exist_ok:
                pass
            else:
                raise FileExistsError
        else:
            s3_client.put_object(
                Bucket=self.bucket,
                Key=self.key,
                Body="",
            )

        if parents:
            for p in self.parents:
                if p.is_bucket() is False:
                    p.mkdir(exist_ok=True, parents=False, bsm=bsm)



# -*- coding: utf-8 -*-

"""
Join path operator implementation.
"""

import typing as T

from .relative import RelativePathAPIMixin
from ..marker import warn_deprecate

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class JoinPathAPIMixin:
    """
    A mixin class that implements the join path operator.
    """

    def join_path(self: "S3Path", *others: "S3Path") -> "S3Path":
        """
        Join with other relative path to form a new path

        Example::

            # create some s3path
            >>> p1 = S3Path("bucket", "folder", "subfolder", "file.txt")
            >>> p2 = p1.parent
            >>> relpath1 = p1.relative_to(p2)

            # preview value
            >>> p1
            S3Path('s3://bucket/folder/subfolder/file.txt')
            >>> p2
            S3Path('s3://bucket/folder/subfolder/')
            >>> relpath1
            S3Path('file.txt')

            # join one relative path
            >>> p2.join_path(relpath1)
            S3Path('s3://bucket/folder/subfolder/file.txt')

            # join multiple relative path
            >>> p3 = p2.parent
            >>> relpath2 = p2.relative_to(p3)
            >>> p3.join_path(relpath2, relpath1)
            S3Path('s3://bucket/folder/subfolder/file.txt')

        :param others: many relative path

        :return: a new :class:`S3Path`

        .. versionadded:: 1.0.1
        """
        warn_deprecate(
            func_name="S3Path.join_path",
            version="2.1.1",
            message="S3Path.joinpath",
        )
        args = [
            self,
        ]
        for relp in others:
            if relp.is_relpath() is False:
                msg = (
                    "you can only join with relative path! " "{} is not a relative path"
                ).format(relp)
                raise TypeError(msg)
            args.append(relp)
        return self._from_parts(args)

    def joinpath(self: "S3Path", *other: T.Union[str, "S3Path"]) -> "S3Path":
        """
        Join with other relative path or string parts.

        Example::

            # join with string parts
            >>> p = S3Path("bucket")
            >>> p.joinpath("folder", "file.txt")
            S3Path('s3://bucket/folder/file.txt')

            # join ith relative path or string parts
            >>> p = S3Path("bucket")
            >>> relpath = S3Path("my-bucket", "data", "folder/").relative_to(S3Path("my-bucket", "data"))
            >>> p.joinpath("data", relpath, "file.txt")
            S3Path('s3://bucket/data/folder/file.txt')

        :param others: many string or relative path

        .. versionadded:: 1.1.1
        """
        args = [
            self,
        ]
        for part in other:
            if isinstance(part, str):
                args.append(part)
            elif isinstance(part, RelativePathAPIMixin):
                if part.is_relpath() is False:
                    msg = (
                        "you can only join with string part or relative path! "
                        "{} is not a relative path"
                    ).format(part)
                    raise TypeError(msg)
                else:
                    args.append(part)
            else:
                msg = (
                    "you can only join with string part or relative path! "
                    "{} is not a relative path"
                ).format(part)
                raise TypeError(msg)
        return self._from_parts(args)

    def __truediv__(
        self: "S3Path", other: T.Union[str, "S3Path", T.List[T.Union[str, "S3Path"]]]
    ) -> "S3Path":
        """
        A syntax sugar. Basically ``S3Path(s3path, part1, part2, ...)``
        is equal to ``s3path / part1 / part2 / ...`` or
        ``s3path / [part1, part2]``

        Example::

            >>> S3Path("bucket") / "folder" / "file.txt"
            S3Path('s3://bucket/folder/file.txt')

            >>> S3Path("bucket") / ["folder", "file.txt"]
            S3Path('s3://bucket/folder/file.txt')

            # relative path also work
            >>> S3Path("new-bucket") / (S3Path("bucket/file.txt").relative_to(S3Path("bucket")))
            S3Path('s3://new-bucket/file.txt')

        .. versionadded:: 1.0.11
        """
        if self.is_void():
            raise TypeError("You cannot do ``VoidS3Path / other``!")
        if isinstance(other, list):
            res = self
            for part in other:
                res = res / part
            return res
        else:
            if (
                self.is_relpath()
                and isinstance(other, RelativePathAPIMixin)
                and other.is_relpath() is False
            ):
                raise TypeError("you cannot do ``RelativeS3Path / NonRelativeS3Path``!")
            return self.joinpath(other)



# -*- coding: utf-8 -*-

import sys
from pathlib import Path

dir_project_root = Path(__file__).absolute().parent.parent.parent

# code structure
dir_tests: Path = dir_project_root / "tests"
dir_htmlcov: Path = dir_project_root / "htmlcov"

# virtual environment
dir_bin: Path = Path(sys.executable).parent
bin_pytest: Path = dir_bin / "pytest"



# -*- coding: utf-8 -*-

"""
Type hint variables.

Note:

- This module is not for public API
"""

import typing as T

from pathlib import Path as Path1
from pathlib_mate import Path as Path2

TagType = T.Dict[str, str]
TagSetType = T.List[T.Dict[str, str]]

MetadataType = T.Dict[str, str]
PathType = T.Union[str, Path1, Path2]



# -*- coding: utf-8 -*-

"""
S3 botocore client wrapper.

Note:

- This module is not for public API
"""

import typing as T
from datetime import datetime

import botocore.exceptions
from func_args import NOTHING, resolve_kwargs

from .type import TagType, MetadataType
from .exc import S3ObjectNotExist
from .tag import parse_tags, encode_tag_set, encode_url_query


def is_bucket_exists(
    s3_client,
    bucket: str,
) -> bool:
    """
    Use head_bucket() api to check if an S3 bucket exists.

    wrapper of `head_bucket <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.head_bucket>`_
    """
    try:
        s3_client.head_bucket(Bucket=bucket)
        return True
    except botocore.exceptions.ClientError as e:
        if "Not Found" in str(e):
            return False
        else:  # pragma: no cover
            raise e


def head_object(
    s3_client,
    bucket: str,
    key: str,
    if_match: str = NOTHING,
    if_modified_since: datetime = NOTHING,
    if_none_match: str = NOTHING,
    if_unmodified_since: datetime = NOTHING,
    range: str = NOTHING,
    version_id: str = NOTHING,
    sse_customer_algorithm: str = NOTHING,
    sse_customer_key: str = NOTHING,
    request_payer: str = NOTHING,
    part_number: int = NOTHING,
    expected_bucket_owner: str = NOTHING,
    checksum_mode: str = NOTHING,
) -> dict:
    """
    Use head_object() api to return metadata of an object.

    wrapper of `head_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.head_object>`_
    """
    try:
        dct = s3_client.head_object(
            **resolve_kwargs(
                Bucket=bucket,
                Key=key,
                IfMatch=if_match,
                IfModifiedSince=if_modified_since,
                IfNoneMatch=if_none_match,
                IfUnmodifiedSince=if_unmodified_since,
                Range=range,
                VersionId=version_id,
                SSECustomerAlgorithm=sse_customer_algorithm,
                SSECustomerKey=sse_customer_key,
                RequestPayer=request_payer,
                PartNumber=part_number,
                ExpectedBucketOwner=expected_bucket_owner,
                ChecksumMode=checksum_mode,
            )
        )
        del dct["ResponseMetadata"]
        return dct
    except botocore.exceptions.ClientError as e:
        if "Not Found" in str(e):
            raise S3ObjectNotExist(str(e))
        else:  # pragma: no cover
            raise e


def head_object_or_none(
    s3_client,
    bucket: str,
    key: str,
) -> T.Optional[dict]:
    """
    Use head_object() api to return metadata of an object.

    Behavior:

    1. return ``dict`` head_object() api response if the object exists.
    2. return ``None`` if object does not exist.
    3. raise exception if other error raised.
    """
    try:
        return head_object(s3_client, bucket, key)
    except S3ObjectNotExist:
        return None


def put_object(
    s3_client,
    bucket: str,
    key: str,
    body: bytes = NOTHING,
    metadata: MetadataType = NOTHING,
    tags: TagType = NOTHING,
    acl: str = NOTHING,
    cache_control: str = NOTHING,
    content_disposition: str = NOTHING,
    content_encoding: str = NOTHING,
    content_language: str = NOTHING,
    content_length: int = NOTHING,
    content_md5: str = NOTHING,
    content_type: str = NOTHING,
    checksum_algorithm: str = NOTHING,
    checksum_crc32: str = NOTHING,
    checksum_crc32c: str = NOTHING,
    checksum_sha1: str = NOTHING,
    checksum_sha256: str = NOTHING,
    expires_datetime: datetime = NOTHING,
    grant_full_control: str = NOTHING,
    grant_read: str = NOTHING,
    grant_read_acp: str = NOTHING,
    grant_write_acp: str = NOTHING,
    server_side_encryption: str = NOTHING,
    storage_class: str = NOTHING,
    website_redirect_location: str = NOTHING,
    sse_customer_algorithm: str = NOTHING,
    sse_customer_key: str = NOTHING,
    sse_kms_key_id: str = NOTHING,
    sse_kms_encryption_context: str = NOTHING,
    bucket_key_enabled: bool = NOTHING,
    request_payer: str = NOTHING,
    object_lock_mode: str = NOTHING,
    object_lock_retain_until_datetime: datetime = NOTHING,
    object_lock_legal_hold_status: str = NOTHING,
    expected_bucket_owner: str = NOTHING,
) -> dict:
    """
    wrapper of `put_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object>`_

    .. versionchanged:: 1.4.1

        add all of boto3 argument.
    """
    return s3_client.put_object(
        **resolve_kwargs(
            Bucket=bucket,
            Key=key,
            Body=body,
            Metadata=metadata,
            Tagging=tags if tags is NOTHING else encode_url_query(tags),
            ACL=acl,
            CacheControl=cache_control,
            ContentDisposition=content_disposition,
            ContentEncoding=content_encoding,
            ContentLanguage=content_language,
            ContentLength=content_length,
            ContentMD5=content_md5,
            ContentType=content_type,
            ChecksumAlgorithm=checksum_algorithm,
            ChecksumCRC32=checksum_crc32,
            ChecksumCRC32C=checksum_crc32c,
            ChecksumSHA1=checksum_sha1,
            ChecksumSHA256=checksum_sha256,
            Expires=expires_datetime,
            GrantFullControl=grant_full_control,
            GrantRead=grant_read,
            GrantReadACP=grant_read_acp,
            GrantWriteACP=grant_write_acp,
            ServerSideEncryption=server_side_encryption,
            StorageClass=storage_class,
            WebsiteRedirectLocation=website_redirect_location,
            SSECustomerAlgorithm=sse_customer_algorithm,
            SSECustomerKey=sse_customer_key,
            SSEKMSKeyId=sse_kms_key_id,
            SSEKMSEncryptionContext=sse_kms_encryption_context,
            BucketKeyEnabled=bucket_key_enabled,
            RequestPayer=request_payer,
            ObjectLockMode=object_lock_mode,
            ObjectLockRetainUntilDate=object_lock_retain_until_datetime,
            ObjectLockLegalHoldStatus=object_lock_legal_hold_status,
            ExpectedBucketOwner=expected_bucket_owner,
        )
    )


def get_object_tagging(
    s3_client,
    bucket: str,
    key: str,
) -> TagType:
    """
    wrapper of `get_object_tagging <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.get_object_tagging>`_
    """
    kwargs = dict(
        Bucket=bucket,
        Key=key,
    )
    try:
        response = s3_client.get_object_tagging(**kwargs)
        return parse_tags(response["TagSet"])
    except Exception as e:
        if "The specified key does not exist" in str(e):
            raise S3ObjectNotExist(f"s3://{bucket}/{key}")
        else:  # pragma: no cover
            raise e


def put_object_tagging(
    s3_client,
    bucket: str,
    key: str,
    tags: TagType,
):
    """
    wrapper of `put_object_tagging <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.get_object_tagging>`_
    """
    kwargs = dict(
        Bucket=bucket,
        Key=key,
        Tagging={"TagSet": encode_tag_set(tags)},
    )
    return s3_client.put_object_tagging(**kwargs)


def update_object_tagging(
    s3_client,
    bucket: str,
    key: str,
    tags: TagType,
) -> TagType:
    """
    Allow you to use ``dict.update`` liked API to update s3 object tagging.
    """
    existing_tags = get_object_tagging(s3_client, bucket, key)
    existing_tags.update(tags)
    put_object_tagging(s3_client, bucket, key, existing_tags)
    return existing_tags


def copy_object(
    s3_client,
    src_bucket: str,
    src_key: str,
    dst_bucket: str,
    dst_key: str,
    metadata: T.Optional[MetadataType] = NOTHING,
    tags: T.Optional[TagType] = NOTHING,
    acl: str = NOTHING,
    cache_control: str = NOTHING,
    content_disposition: str = NOTHING,
    content_encoding: str = NOTHING,
    content_language: str = NOTHING,
    content_md5: str = NOTHING,
    content_type: str = NOTHING,
    copy_source_if_match: str = NOTHING,
    copy_source_if_modified_since: datetime = NOTHING,
    copy_source_if_none_match: str = NOTHING,
    copy_source_if_unmodified_since: datetime = NOTHING,
    expires_datetime: datetime = NOTHING,
    grant_full_control: str = NOTHING,
    grant_read: str = NOTHING,
    grant_read_acp: str = NOTHING,
    grant_write_acp: str = NOTHING,
    server_side_encryption: str = NOTHING,
    storage_class: str = NOTHING,
    website_redirect_location: str = NOTHING,
    sse_customer_algorithm: str = NOTHING,
    sse_customer_key: str = NOTHING,
    sse_kms_key_id: str = NOTHING,
    sse_kms_encryption_context: str = NOTHING,
    bucket_key_enabled: bool = NOTHING,
    copy_source_sse_customer_algorithm: str = NOTHING,
    copy_source_sse_customer_key: str = NOTHING,
    request_payer: str = NOTHING,
    object_lock_mode: str = NOTHING,
    object_lock_retain_until_datetime: datetime = NOTHING,
    object_lock_legal_hold_status: str = NOTHING,
    expected_bucket_owner: str = NOTHING,
    expected_source_bucket_owner: str = NOTHING,
) -> dict:
    """
    wrapper of `copy_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.copy_object>`_
    """
    kwargs = dict(
        Bucket=dst_bucket,
        Key=dst_key,
        CopySource=dict(
            Bucket=src_bucket,
            Key=src_key,
        ),
    )
    if metadata is not NOTHING:
        kwargs["Metadata"] = metadata
        kwargs["MetadataDirective"] = "REPLACE"
    if tags is not NOTHING:
        kwargs["Tagging"] = encode_url_query(tags)
        kwargs["TaggingDirective"] = "REPLACE"

    kwargs.update(
        dict(
            ACL=acl,
            CacheControl=cache_control,
            ContentDisposition=content_disposition,
            ContentEncoding=content_encoding,
            ContentLanguage=content_language,
            ContentMD5=content_md5,
            ContentType=content_type,
            CopySourceIfMatch=copy_source_if_match,
            CopySourceIfModifiedSince=copy_source_if_modified_since,
            CopySourceIfNoneMatch=copy_source_if_none_match,
            CopySourceIfUnmodifiedSince=copy_source_if_unmodified_since,
            Expires=expires_datetime,
            GrantFullControl=grant_full_control,
            GrantRead=grant_read,
            GrantReadACP=grant_read_acp,
            GrantWriteACP=grant_write_acp,
            ServerSideEncryption=server_side_encryption,
            StorageClass=storage_class,
            WebsiteRedirectLocation=website_redirect_location,
            SSECustomerAlgorithm=sse_customer_algorithm,
            SSECustomerKey=sse_customer_key,
            SSEKMSKeyId=sse_kms_key_id,
            SSEKMSEncryptionContext=sse_kms_encryption_context,
            BucketKeyEnabled=bucket_key_enabled,
            CopySourceSSECustomerAlgorithm=copy_source_sse_customer_algorithm,
            CopySourceSSECustomerKey=copy_source_sse_customer_key,
            RequestPayer=request_payer,
            ObjectLockMode=object_lock_mode,
            ObjectLockRetainUntilDate=object_lock_retain_until_datetime,
            ObjectLockLegalHoldStatus=object_lock_legal_hold_status,
            ExpectedBucketOwner=expected_bucket_owner,
            ExpectedSourceBucketOwner=expected_source_bucket_owner,
        )
    )
    return s3_client.copy_object(**resolve_kwargs(**kwargs))



# -*- coding: utf-8 -*-

"""
S3Path property methods.
"""

import typing as T

from .filterable_property import FilterableProperty

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class AttributeAPIMixin:
    """
    A mixin class that implements the property methods.
    """

    @property
    def parent(self: "S3Path") -> T.Optional["S3Path"]:
        """
        Return the parent s3 directory.

        - if current object is on s3 bucket root directory, it returns bucket
            root directory
        - it is always a directory (``s3path.is_dir() is True``)
        - if it is already s3 bucket root directory, it returns ``None``

        Examples::

            >>> S3Path("my-bucket", "my-folder", "my-file.json").parent.uri
            s3://my-bucket/my-folder/

            >>> S3Path("my-bucket", "my-folder", "my-subfolder/").parent.uri
            s3://my-bucket/my-folder/

            >>> S3Path("my-bucket", "my-folder").parent.uri
            s3://my-bucket/

            >>> S3Path("my-bucket", "my-file.json").parent.uri
            s3://my-bucket/

        .. versionadded:: 1.0.1
        """
        if len(self._parts) == 0:
            return self
        else:
            return self._from_parsed_parts(
                bucket=self._bucket,
                parts=self._parts[:-1],
                is_dir=True,
            )

    @property
    def parents(self: "S3Path") -> T.List["S3Path"]:
        """
        An immutable sequence providing access to the logical ancestors of
        the path.

        .. versionadded:: 1.0.6
        """
        if self.is_void():
            raise ValueError(f"void S3path doesn't support .parents method!")
        if self.is_relpath():
            raise ValueError(f"relative S3path doesn't support .parents method!")
        l = list()
        parent = self
        while 1:
            new_parent = parent.parent
            if parent.uri == new_parent.uri:
                break
            else:
                l.append(new_parent)
                parent = new_parent
        return l

    def is_parent_of(self: "S3Path", other: "S3Path") -> bool:
        """
        Test if it is the parent directory or grand-grand-... parent directory
        of another one.

        Examples::

            # is parent
            >>> S3Path("bucket").is_parent_of(S3Path("bucket/folder/"))
            True

            # is grand parent
            >>> S3Path("bucket").is_parent_of(S3Path("bucket/folder/file.txt"))
            True

            # the root bucket's parent is itself
            >>> S3Path("bucket").is_parent_of(S3Path("bucket"))
            True

            # the 'self' has to be a directory
            >>> S3Path("bucket/a").is_parent_of(S3Path("bucket/a/b/c"))
            TypeError: S3Path('s3://bucket/a') is not a valid directory!

            # the 'self' and 'other' has to be concrete S3Path
            >>> S3Path().is_parent_of(S3Path())
            TypeError: both S3Path(), S3Path() has to be a concrete S3Path!

        .. versionadded:: 1.0.2
        """
        if self._bucket is None or other._bucket is None:
            raise TypeError(f"both {self}, {other} has to be a concrete S3Path!")
        if self._bucket != other._bucket:
            return False
        if self.is_dir() is False:
            raise TypeError(f"{self} is not a valid directory!")
        n_parts_other = len(other.parts)
        if n_parts_other == 0:
            return len(self._parts) == 0
        else:
            return (
                self._parts[: (n_parts_other - 1)] == other._parts[:-1]
                and len(self._parts) < n_parts_other
            )

    def is_prefix_of(self: "S3Path", other: "S3Path") -> bool:
        """
        Test if it is a prefix of another one.

        Example::

            >>> S3Path("bucket/folder/").is_prefix_of(S3Path("bucket/folder/file.txt"))
            True

            >>> S3Path("bucket/folder/").is_prefix_of(S3Path("bucket/folder/"))
            True

        .. versionadded:: 1.0.2
        """
        if self._bucket is None or other._bucket is None:
            raise TypeError(f"both {self}, {other} has to be a concrete S3Path!")
        if self._bucket != other._bucket:
            return False
        return self.uri <= other.uri

    @FilterableProperty
    def basename(self: "S3Path") -> T.Optional[str]:
        """
        The file name with extension, or the last folder name if it is a
        directory. If not available, it returns None. For example it doesn't
        make sence for s3 bucket.

        Logically: dirname + basename = abspath

        Example::

            # s3 object
            >>> S3Path("bucket", "folder", "file.txt").basename
            'file.txt'

            # s3 directory
            >>> S3Path("bucket", "folder/").basename
            'folder'

            # s3 bucket
            >>> S3Path("bucket").basename
            None

            # void path
            >>> S3Path().basename
            ''

            # relative path
            >>> S3Path.make_relpath("folder", "file.txt").basename
            None

        .. versionadded:: 1.0.1
        """
        if len(self._parts):
            return self._parts[-1]
        else:
            return ""

    @FilterableProperty
    def dirname(self: "S3Path") -> T.Optional[str]:
        """
        The basename of it's parent directory.

        Example::

            >>> S3Path("bucket", "folder", "file.txt").dirname
            'folder'

            # root dir name is ''
            >>> S3Path("bucket", "folder").dirname
            ''

        .. versionadded:: 1.0.1
        """
        return self.parent.basename

    @FilterableProperty
    def fname(self: "S3Path") -> str:
        """
        The final path component, minus its last suffix (file extension).
        Only if it is not a directory.

        Example::

            >>> S3Path("bucket", "folder", "file.txt").fname
            'file'

        .. versionadded:: 1.0.1
        """
        if self.is_dir():
            raise TypeError
        basename: str = self.basename
        if not basename:
            raise ValueError
        i = basename.rfind(".")
        if 0 < i < len(basename) - 1:
            return basename[:i]
        else:
            return basename

    @FilterableProperty
    def ext(self: "S3Path") -> str:
        """
        The final component's last suffix, if any. Usually it is the file
        extension. Only if it is not a directory.

        Example::

            >>> S3Path("bucket", "folder", "file.txt").fname
            '.txt'

        .. versionadded:: 1.0.1
        """
        if self.is_dir():
            raise TypeError
        basename: str = self.basename
        if not basename:
            raise ValueError
        i = basename.rfind(".")
        if 0 < i < len(basename) - 1:
            return basename[i:]
        else:
            return ""

    @FilterableProperty
    def abspath(self: "S3Path") -> str:
        """
        The Unix styled absolute path from the bucket. You can think of the
        bucket as a root drive.

        Example::

            # s3 object
            >>> S3Path("bucket", "folder", "file.txt").abspath
            '/folder/file.txt'

            # s3 directory
            >>> S3Path("bucket", "folder/").abspath
            '/folder/'

            # s3 bucket
            >>> S3Path("bucket").abspath
            '/'

            # void path
            >>> S3Path().abspath
            TypeError: relative path doesn't have absolute path!

            # relative path
            >>> S3Path.make_relpath("folder", "file.txt").abspath
            TypeError: relative path doesn't have absolute path!

        .. versionadded:: 1.0.1
        """
        if self._bucket is None:
            raise TypeError("relative path doesn't have absolute path!")
        if self.is_bucket():
            return "/"
        elif self.is_dir():
            return "/" + "/".join(self._parts) + "/"
        elif self.is_file():
            return "/" + "/".join(self._parts)
        else:  # pragma: no cover
            raise TypeError

    @FilterableProperty
    def dirpath(self: "S3Path"):
        """
        The Unix styled absolute path from the bucket of the **parent directory**.

        Example::

            # s3 object
            >>> S3Path("bucket", "folder", "file.txt").dirpath
            '/folder/'

            # s3 directory
            >>> S3Path("bucket", "folder/").dirpath
            '/'

            # s3 bucket
            >>> S3Path("bucket").dirpath
            '/'

            # void path
            >>> S3Path().dirpath
            TypeError: relative path doesn't have absolute path!

            # relative path
            >>> S3Path.make_relpath("folder", "file.txt").dirpath
            TypeError: relative path doesn't have absolute path!

        .. versionadded:: 1.0.2
        """
        return self.parent.abspath

    @property
    def root(self: "S3Path") -> "S3Path":
        """
        Return the S3 Bucket root folder S3Path object.

        Example::

            # s3 object
            >>> S3Path("bucket", "folder", "file.txt").root
            '/folder/'
        """
        if self.is_relpath() or self.is_void():
            raise TypeError("only concrete File or Directory has a bucket root!")
        else:
            return self._from_parsed_parts(
                bucket=self.bucket,
                parts=[],
                is_dir=True,
            )

    def __repr__(self: "S3Path"):
        """
        You cannot use the returned string of __repr__ to recover the original
        S3Path method.
        """
        classname = self.__class__.__name__

        if self.is_void():
            return "S3VoidPath()"

        elif self.is_relpath():
            key = self.key
            if len(key):
                return f"S3RelPath({key!r})"
            else:  # pragma: no cover
                return "S3RelPath()"
        else:
            uri = self.uri
            return "{}('{}')".format(classname, uri)

    def __str__(self: "S3Path"):
        return self.__repr__()



# -*- coding: utf-8 -*-

"""
Sync file, folder between s3-to-s3, s3-to-local, local-to-s3.
"""

import typing as T

import subprocess

from boto_session_manager import BotoSesManager

from ..aws import context
from ..type import PathType

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


def _preprocess(path: T.Union["S3Path", PathType]) -> T.Union["S3Path", str]:
    if hasattr(path, "console_url"):
        return path
    else:
        return str(path)


class SyncAPIMixin:
    """
    A mixin class that implements aws s3 sync feature.
    """

    @classmethod
    def sync(
        cls: T.Type["S3Path"],
        src: T.Union["S3Path", PathType],
        dst: T.Union["S3Path", PathType],
        bsm: T.Optional["BotoSesManager"] = None,
        quite: bool = True,
        include: T.Optional[str] = None,
        exclude: T.Optional[str] = None,
        acl: T.Optional[str] = None,
        only_show_errors: bool = False,
        no_progress: bool = False,
        page_size: T.Optional[str] = None,
        delete: bool = False,
    ):
        """
        Implement the `aws s3 sync <https://docs.aws.amazon.com/cli/latest/reference/s3/sync.html>`_
        CLI behavior.

        :param src:
        :param dst:

        .. versionadded:: 1.2.1

        TODO: add support for all aws s3 sync supported arguments
        """
        args = [
            "aws",
            "s3",
            "sync",
        ]

        # handle source and target
        src = _preprocess(src)
        dst = _preprocess(dst)

        if isinstance(src, str):
            if src.startswith("s3://"):
                src_is_s3 = True
            else:
                src_is_s3 = False
            src_arg = src.replace("\\", "/")
        else:
            src_is_s3 = True
            src_arg = src.uri

        if isinstance(dst, str):
            if dst.startswith("s3://"):
                dst_is_s3 = True
            else:
                dst_is_s3 = False
            dst_arg = dst.replace("\\", "/")
        else:
            dst_is_s3 = True
            dst_arg = dst.uri

        if src_is_s3 is False and dst_is_s3 is False:
            raise ValueError

        args.extend([src_arg, dst_arg])

        # handle additional CLI argument
        if quite:  # pragma: no cover
            args.append("--quiet")
        if include:  # pragma: no cover
            args.extend(["--include", include])
        if exclude:  # pragma: no cover
            args.extend(["--exclude", exclude])
        if acl:  # pragma: no cover
            args.extend(["--acl", acl])
        if only_show_errors:  # pragma: no cover
            args.append("--only-show-errors")
        if no_progress:  # pragma: no cover
            args.append("--no-progress")
        if page_size:  # pragma: no cover
            args.extend(["--page-size", str(page_size)])
        if delete:  # pragma: no cover
            args.append("--delete")

        if bsm is None:  # pragma: no cover
            with BotoSesManager(botocore_session=context.boto_ses._session).awscli():
                print(" ".join(args))
                response = subprocess.run(args)
        else:  # pragma: no cover
            with bsm.awscli():
                response = subprocess.run(args)
        if response.returncode != 0:  # pragma: no cover
            raise SystemError("'aws s3 sync' command failed!")

    def sync_from(
        self: "S3Path",
        src: T.Union["S3Path", PathType],
        bsm: T.Optional["BotoSesManager"] = None,
        quite: bool = True,
        include: T.Optional[str] = None,
        exclude: T.Optional[str] = None,
        acl: T.Optional[str] = None,
        only_show_errors: bool = False,
        no_progress: bool = False,
        page_size: T.Optional[str] = None,
        delete: bool = False,
    ):
        """
        Sync data from external place to this S3 location.
        """
        return self.sync(
            src=src,
            dst=self,
            bsm=bsm,
            quite=quite,
            include=include,
            exclude=exclude,
            acl=acl,
            only_show_errors=only_show_errors,
            no_progress=no_progress,
            page_size=page_size,
            delete=delete,
        )

    def sync_to(
        self: "S3Path",
        dst: T.Union["S3Path", PathType],
        bsm: T.Optional["BotoSesManager"] = None,
        quite: bool = True,
        include: T.Optional[str] = None,
        exclude: T.Optional[str] = None,
        acl: T.Optional[str] = None,
        only_show_errors: bool = False,
        no_progress: bool = False,
        page_size: T.Optional[str] = None,
        delete: bool = False,
    ):
        """
        Sync the data at this S3 location to external place.
        """
        return self.sync(
            src=self,
            dst=dst,
            bsm=bsm,
            quite=quite,
            include=include,
            exclude=exclude,
            acl=acl,
            only_show_errors=only_show_errors,
            no_progress=no_progress,
            page_size=page_size,
            delete=delete,
        )



# -*- coding: utf-8 -*-

"""
serialization and deserialization.
"""

import typing as T

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class SerdeAPIMixin:
    """
    A mixin class that implements the serialization and deserialization.
    """

    def to_dict(self: "S3Path") -> dict:
        """
        Serialize to Python dict

        .. versionadded:: 1.0.1
        """
        return {
            "bucket": self._bucket,
            "parts": self._parts,
            "is_dir": self._is_dir,
        }

    @classmethod
    def from_dict(cls: "S3Path", dct: dict) -> "S3Path":
        """
        Deserialize from Python dict

        .. versionadded:: 1.0.2
        """
        return cls._from_parsed_parts(
            bucket=dct["bucket"],
            parts=dct["parts"],
            is_dir=dct["is_dir"],
        )



# -*- coding: utf-8 -*-

"""
The base constructor of S3Path object.
"""

import typing as T

try:
    import botocore.exceptions
except ImportError:  # pragma: no cover
    pass
except:  # pragma: no cover
    raise

from .filterable_property import FilterableProperty
from .. import utils

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class BaseS3Path:
    """
    Similar to ``pathlib.Path``. An objective oriented programming interface
    for AWS S3 object or logical directory.

    You can use this class in different way.

    1. pure s3 object / directory path manipulation without actually
        talking to AWS API.
    2. get metadata of an object, count objects, get statistics information
        of a directory
    3. enhanced s3 API that do: :meth:`upload_file <upload_file>`,
        :meth:`upload_dir <upload_dir>`,
        :meth:`copy <copy_to>` a file or directory,
        :meth:`move <move_to>` a file or directory,
        :meth:`delete <delete_if_exists>`  a file or directory,
        :meth:`iter_objects <iter_objects>` from a prefix.

    **Constructor**

    The :class:`S3Path` itself is a constructor. It takes ``str`` and other
    relative :class:`S3Path` as arguments.

    1. The first argument defines the S3 bucket
    2. The rest of arguments defines the path parts of the S3 key
    3. The final argument defines the type whether is a file (object) or
        a directory

    First let's create a S3 object path from string::

        # first argument becomes the bucket
        >>> s3path = S3Path("bucket", "folder", "file.txt")
        # print S3Path object gives you info in S3 URI format
        >>> s3path
        S3Path('s3://bucket/folder/file.txt')

        # last argument defines that it is a file
        >>> s3path.is_file()
        True
        >>> s3path.is_dir()
        True

        # "/" separator will be automatically handled
        >>> S3Path("bucket", "folder/file.txt")
        S3Path('s3://bucket/folder/file.txt')

        >>> S3Path("bucket/folder/file.txt")
        S3Path('s3://bucket/folder/file.txt')

    Then let's create a S3 directory path::

        >>> s3path= S3Path("bucket/folder/")
        >>> s3path
        S3Path('s3://bucket/folder/')

        # last argument defines that it is a directory
        >>> s3path.is_dir()
        True
        >>> s3path.is_file()
        False

    .. versionadded:: 1.0.1
    """

    __slots__ = (
        "_bucket",
        "_parts",
        "_is_dir",
        "_cached_cparts",  # cached comparison parts
        "_hash",  # cached hash value
        "_meta",  # s3 object metadata cache object
    )

    def __new__(
        cls: T.Type["S3Path"],
        *args: T.Union[str, "S3Path"],
    ) -> "S3Path":
        return cls._from_parts(args)

    @classmethod
    def _from_parts(
        cls: T.Type["S3Path"],
        args: T.List[T.Union[str, "S3Path"]],
        init: bool = True,
    ) -> "S3Path":
        _bucket = None
        _parts = list()
        _is_dir = None

        if len(args) == 0:
            return cls._from_parsed_parts(
                bucket=_bucket,
                parts=_parts,
                is_dir=_is_dir,
            )

        # resolve self._bucket
        arg = args[0]
        if isinstance(arg, str):
            utils.validate_s3_bucket(arg)
            parts = utils.split_parts(arg)
            _bucket = parts[0]
            _parts.extend(parts[1:])
        elif isinstance(arg, BaseS3Path):
            _bucket = arg._bucket
            _parts.extend(arg._parts)
        else:
            raise TypeError

        # resolve self._parts
        for arg in args[1:]:
            if isinstance(arg, str):
                utils.validate_s3_key(arg)
                _parts.extend(utils.split_parts(arg))
            elif isinstance(arg, BaseS3Path):
                if arg._bucket is None:
                    _parts.extend(arg._parts)
                else:
                    raise TypeError(
                        "from the second arguments, it has to be raw string "
                        "(as a part) or a relative S3Path (without bucket)! "
                        f"this is invalid: {arg}."
                    )
            else:
                raise TypeError

        # resolve self._is_dir
        # inspect the last argument
        if isinstance(arg, str):
            _is_dir = arg.endswith("/")
        elif isinstance(arg, BaseS3Path):
            _is_dir = arg._is_dir
        else:  # pragma: no cover
            raise TypeError

        if (_bucket is not None) and len(_parts) == 0:  # bucket root
            _is_dir = True

        return cls._from_parsed_parts(
            bucket=_bucket,
            parts=_parts,
            is_dir=_is_dir,
            init=init,
        )

    @classmethod
    def _from_parsed_parts(
        cls: T.Type["S3Path"],
        bucket: T.Optional[str],
        parts: T.List[str],
        is_dir: T.Optional[bool],
        init: bool = True,
    ) -> "S3Path":
        self = object.__new__(cls)
        self._bucket = bucket
        self._parts = parts
        self._is_dir = is_dir
        self._meta = None
        if init:
            self._init()
        return self

    def _init(self: "S3Path") -> None:
        """
        Additional instance initialization
        """
        pass

    @FilterableProperty
    def parts(self: "S3Path") -> T.List[str]:
        """
        Provides sequence-like access to the components in the filesystem path.
        It doesn't include the bucket, because bucket is considered as "drive".

        .. versionadded:: 1.0.1
        """
        return self._parts



# -*- coding: utf-8 -*-

"""
S3Path object mutation implementation.
"""

import typing as T

from .. import exc

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class MutateAPIMixin:
    """
    A mixin class that implements the S3Path object mutation.
    """

    def copy(self: "S3Path") -> "S3Path":
        """
        Create a copy of S3Path object that logically equals to this one,
        but is actually different identity in memory. Also, the cache data
        are cleared.

        Example::

            >>> p1 = S3Path("bucket", "folder", "file.txt")
            >>> p2 = p1.copy()
            >>> p1 == p2
            True
            >>> p1 is p2
            False

        .. versionadded:: 1.0.1
        """
        return self._from_parsed_parts(
            bucket=self._bucket,
            parts=list(self._parts),
            is_dir=self._is_dir,
        )

    def change(
        self: "S3Path",
        new_bucket: str = None,
        new_abspath: str = None,
        new_dirpath: str = None,
        new_dirname: str = None,
        new_basename: str = None,
        new_fname: str = None,
        new_ext: str = None,
    ) -> "S3Path":
        """
        Create a new S3Path by replacing part of the attributes. If no argument
        is given, it behaves like :meth:`copy`.

        :param new_bucket: The new bucket name
        :param new_abspath:
        :param new_dirpath:
        :param new_dirname:
        :param new_basename:
        :param new_fname:
        :param new_ext:

        .. versionadded:: 1.0.2
        """
        if new_bucket is None:
            new_bucket = self.bucket

        if new_abspath is not None:
            exc.ensure_all_none(
                new_dirpath=new_dirpath,
                new_dirname=new_dirname,
                new_basename=new_basename,
                new_fname=new_fname,
                new_ext=new_ext,
            )
            p = self._from_parts([self.bucket, new_abspath])
            return p

        if (new_dirpath is None) and (new_dirname is not None):
            dir_parts = self.parent.parent._parts + [new_dirname]
        elif (new_dirpath is not None) and (new_dirname is None):
            dir_parts = [
                new_dirpath,
            ]
        elif (new_dirpath is None) and (new_dirname is None):
            dir_parts = self.parent._parts
        else:
            raise ValueError("Cannot having both 'new_dirpath' and 'new_dirname'!")

        if new_basename is None:
            if new_fname is None:
                new_fname = self.fname
            if new_ext is None:
                new_ext = self.ext
            new_basename = new_fname + new_ext
        else:
            if (new_fname is not None) or (new_ext is not None):
                raise ValueError(
                    "Cannot having both "
                    "'new_basename' / 'new_fname', "
                    "or 'new_basename' / 'new_ext'!"
                )
        if new_bucket is None:
            p = self._from_parts(
                [
                    "dummy-bucket",
                ]
                + dir_parts
                + [
                    new_basename,
                ]
            )
            p._bucket = None
        else:
            p = self._from_parts(
                [
                    new_bucket,
                ]
                + dir_parts
                + [
                    new_basename,
                ]
            )
        return p

    def to_dir(self: "S3Path") -> "S3Path":
        if self.is_dir():
            return self.copy()
        elif self.is_file():
            return self.joinpath("/")
        else:
            raise ValueError("only concrete file or folder S3Path can do .to_dir()")

    def to_file(self: "S3Path") -> "S3Path":
        if self.is_file():
            return self.copy()
        elif self.is_dir():
            p = self.copy()
            p._is_dir = False
            return p
        else:
            raise ValueError("only concrete file or folder S3Path can do .to_file()")



# -*- coding: utf-8 -*-

"""
This module provides AWS Tags manipulation helpers.

Note:

- This module is not for public API
"""

import typing as T
from urllib.parse import urlencode
from .type import TagType, TagSetType


def parse_tags(
    data: T.Union[
        TagSetType,
    ]
) -> TagType:
    """
    Convert tags to pythonic dictionary.
    """
    if isinstance(data, list):
        return {dct["Key"]: dct["Value"] for dct in data}
    else:  # pragma: no cover
        raise NotImplementedError


def encode_tag_set(tags: TagType) -> TagSetType:
    """
    Some API requires: ``[{"Key": "name", "Value": "Alice"}, {...}, ...]``
    for tagging parameter.
    """
    return [
        {"Key": k, "Value": v}
        for k, v in tags.items()
    ]


def encode_url_query(tags: TagType) -> str:
    """
    Some API requires: ``Key1=Value1&Key2=Value2`` for tagging parameter.
    """
    return urlencode(tags)



# -*- coding: utf-8 -*-

"""
Tagging related API.
"""

import typing as T
import botocore.exceptions

from .resolve_s3_client import resolve_s3_client
from .. import utils, client as better_client
from ..aws import context

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class ExistsAPIMixin:
    """
    A mixin class that implements the exists test related methods.
    """
    def exists(
        self: "S3Path",
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> bool:
        """
        - For S3 bucket: check if the bucket exists. If you don't have the
            access, then it raise exception.
        - For S3 object: check if the object exists
        - For S3 directory: check if the directory exists, it returns ``True``
            even if the folder doesn't have any object.

        .. versionadded:: 1.0.1
        """
        if self.is_bucket():
            s3_client = resolve_s3_client(context, bsm)
            return better_client.is_bucket_exists(s3_client, self.bucket)
        elif self.is_file():
            s3_client = resolve_s3_client(context, bsm)
            dct = utils.head_object_if_exists(
                s3_client=s3_client,
                bucket=self.bucket,
                key=self.key,
            )
            if isinstance(dct, dict):
                self._meta = dct
                return True
            else:
                return False
        elif self.is_dir():
            l = list(
                self.iter_objects(
                    batch_size=1,
                    limit=1,
                    include_folder=True,
                    bsm=bsm,
                )
            )
            if len(l):
                return True
            else:
                return False
        else:  # pragma: no cover
            raise TypeError

    def ensure_not_exists(
        self: "S3Path",
        bsm: T.Optional["BotoSesManager"] = None,
    ) -> None:
        """
        A validator method ensure that it doesn't exists.

        .. versionadded:: 1.0.1
        """
        if self.exists(bsm=bsm):
            utils.raise_file_exists_error(self.uri)



# -*- coding: utf-8 -*-

"""
S3 URI (Uniform Resource Identifier), ARN (Amazon Resource Name),
console URL related API.
"""

import typing as T

from .filterable_property import FilterableProperty
from .. import utils, validate

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class UriAPIMixin:
    """
    A mixin class that implements the S3 URI, ARN, console url etc ...
    """
    @FilterableProperty
    def bucket(self: 'S3Path') -> T.Optional[str]:
        """
        Return bucket name as string, if available.

        Example::

            >>> S3Path("bucket/folder/file.txt").bucket
            'bucket'

        .. versionadded:: 1.0.1
        """
        return self._bucket

    @FilterableProperty
    def key(self: 'S3Path') -> T.Optional[str]:
        """
        Return object or directory key as string, if available.

        Examples::

            # a s3 object
            >>> S3Path("bucket/folder/file.txt").key
            'folder/file.txt'

            # a s3 object
            >>> S3Path("bucket/folder/").key
            'folder/file.txt'

            # a relative path
            >>> S3Path("bucket/folder/file.txt").relative_to(S3Path("bucket")).key
            'folder/file.txt

            >>> S3Path("bucket/folder/").relative_to(S3Path("bucket")).key
            'folder/'

            # an empty S3Path
            >>> S3Path().key
            ''

        .. versionadded:: 1.0.1
        """
        if len(self._parts):
            return "{}{}".format(
                "/".join(self._parts),
                "/" if self._is_dir else ""
            )
        else:
            return ""

    @FilterableProperty
    def uri(self: 'S3Path') -> T.Optional[str]:
        """
        Return AWS S3 URI.

        - for regular s3 object, it returns ``"s3://{bucket}/{key}"``
        - if it is a directory, the s3 uri always ends with ``"/"``.
        - if it is bucket only (no key), it returns ``"s3://{bucket}/"``
        - if it is not an concrete S3Path, it returns ``None``
        - it has to have bucket, if not (usually because it is an relative path)
            it returns ``None``

        Examples::

            >>> S3Path("bucket", "folder", "file.txt").uri
            's3://bucket/folder/file.txt'

            >>> S3Path("bucket", "folder/").uri
            's3://bucket/folder/'

            >>> S3Path("bucket").uri
            's3://bucket/'

            # void path doesn't have uri
            >>> S3Path().uri
            None

            # relative path doesn't have uri
            >>> S3Path("bucket/folder/file.txt").relative_to(S3Path("bucket")).uri
            None

        .. versionadded:: 1.0.1
        """
        if self._bucket is None:
            return None
        if len(self._parts):
            return "s3://{}/{}".format(
                self.bucket,
                self.key,
            )
        else:
            return "s3://{}/".format(self._bucket)

    @property
    def console_url(self: 'S3Path') -> T.Optional[str]:
        """
        Return an AWS S3 Console url that can inspect the details.

        .. versionadded:: 1.0.1
        """
        uri: str = self.uri
        if uri is None:
            return None
        else:
            console_url = utils.make_s3_console_url(s3_uri=uri)
            return console_url

    @property
    def us_gov_cloud_console_url(self: 'S3Path') -> T.Optional[str]:
        """
        Return an AWS US Gov Cloud S3 Console url that can inspect the details.

        .. versionadded:: 1.0.5
        """
        uri: str = self.uri
        if uri is None:
            return None
        else:
            console_url = utils.make_s3_console_url(
                s3_uri=uri, is_us_gov_cloud=True
            )
            return console_url

    @property
    def s3_select_console_url(self: 'S3Path') -> T.Optional[str]:
        """
        Return an AWS US Gov Cloud S3 Console url that can inspect data with s3 select.

        .. versionadded:: 1.0.12
        """
        if self.is_file():
            return utils.make_s3_select_console_url(
                bucket=self.bucket,
                key=self.key,
                is_us_gov_cloud=False,
            )
        else:
            raise TypeError("you can only do s3 select with an object!")

    @property
    def s3_select_us_gov_cloud_console_url(self: 'S3Path') -> T.Optional[str]:
        """

        Return an AWS S3 Console url that can inspect data with s3 select.

        .. versionadded:: 1.0.12
        """
        if self.is_file():
            return utils.make_s3_select_console_url(
                bucket=self.bucket,
                key=self.key,
                is_us_gov_cloud=True,
            )
        else:
            raise TypeError("you can only do s3 select with an object!")

    @FilterableProperty
    def arn(self: 'S3Path') -> T.Optional[str]:
        """
        Return an AWS S3 Resource ARN. See `ARN definition here <https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html>`_

        .. versionadded:: 1.0.1
        """
        if self._bucket is None:
            return None
        if len(self._parts):
            return "arn:aws:s3:::{}/{}".format(
                self.bucket,
                self.key,
            )
        else:
            return "arn:aws:s3:::{}".format(self._bucket)

    @classmethod
    def from_s3_uri(cls: T.Type['S3Path'], uri: str) -> 'S3Path':
        """
        Construct an :class:`S3Path` from S3 URI.

        >>> p = S3Path.from_s3_uri("s3://bucket/folder/file.txt")

        >>> p
        S3Path('s3://bucket/folder/file.txt')

        >>> p.uri
        's3://bucket/folder/file.txt'
        """
        validate.validate_s3_uri(uri)
        bucket, key = utils.split_s3_uri(uri)
        return cls._from_parts([bucket, key])

    @classmethod
    def from_s3_arn(cls: T.Type['S3Path'], arn: str) -> 'S3Path':
        """

        :param arn:
        :return:
        """
        validate.validate_s3_arn(arn)
        return cls._from_parts([arn.replace("arn:aws:s3:::", "", 1), ])



# -*- coding: utf-8 -*-

"""
Delete S3 file or folder.
"""

import typing as T
from func_args import NOTHING

from .resolve_s3_client import resolve_s3_client
from .. import utils
from ..aws import context

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path
    from boto_session_manager import BotoSesManager


class DeleteAPIMixin:
    """
    A mixin class that implements delete method.
    """

    def delete_if_exists(
        self: "S3Path",
        mfa: str = None,
        version_id: str = None,
        request_payer: str = None,
        bypass_governance_retention: bool = None,
        expected_bucket_owner: str = None,
        include_folder: bool = True,
        bsm: T.Optional["BotoSesManager"] = None,
    ):
        """
        Delete an object or an entire directory. Will do nothing
        if it doesn't exist.

        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.delete_object

        :param include_folder: see :meth:`iter_objects`

        :return: number of object is deleted

        .. versionadded:: 1.0.1
        """
        s3_client = resolve_s3_client(context, bsm)
        if self.is_file():
            if self.exists(bsm=bsm):
                kwargs = dict(
                    Bucket=self.bucket,
                    Key=self.key,
                )
                additional_kwargs = utils.collect_not_null_kwargs(
                    MFA=mfa,
                    VersionId=version_id,
                    RequestPayer=request_payer,
                    BypassGovernanceRetention=bypass_governance_retention,
                    ExpectedBucketOwner=expected_bucket_owner,
                )
                kwargs.update(additional_kwargs)
                s3_client.delete_object(**kwargs)
                return 1
            else:
                return 0
        elif self.is_dir():
            return utils.delete_dir(
                s3_client=s3_client,
                bucket=self.bucket,
                prefix=self.key,
                mfa=mfa,
                request_payer=request_payer,
                bypass_governance_retention=bypass_governance_retention,
                expected_bucket_owner=expected_bucket_owner,
                include_folder=include_folder,
            )
        else:  # pragma: no cover
            raise ValueError



# -*- coding: utf-8 -*-

"""
Comparison operator implementation.
"""

import typing as T

if T.TYPE_CHECKING:  # pragma: no cover
    from .s3path import S3Path


class ComparisonAPIMixin:
    """
    A mixin class that implements the comparison operator magic methods.
    """
    @property
    def _cparts(self: "S3Path"):
        """
        Cached comparison parts, for hashing and comparison
        """
        try:
            return self._cached_cparts
        except AttributeError:
            cparts = list()

            if self._bucket:
                cparts.append(self._bucket)
            else:
                cparts.append("")

            cparts.extend(self._parts)

            if self._is_dir:
                cparts.append("/")

            self._cached_cparts = cparts
            return self._cached_cparts

    def __eq__(self: "S3Path", other: "S3Path") -> bool:
        """
        Return ``self == other``.
        """
        return self._cparts == other._cparts

    def __lt__(self: "S3Path", other: "S3Path") -> bool:
        """
        Return ``self < other``.
        """
        return self._cparts < other._cparts

    def __gt__(self: "S3Path", other: "S3Path") -> bool:
        """
        Return ``self > other``.
        """
        return self._cparts > other._cparts

    def __le__(self: "S3Path", other: "S3Path") -> bool:
        """
        Return ``self <= other``.
        """
        return self._cparts <= other._cparts

    def __ge__(self: "S3Path", other: "S3Path") -> bool:
        """
        Return ``self >= other``.
        """
        return self._cparts >= other._cparts

    def __hash__(self: "S3Path") -> int:
        """
        Return ``hash(self)``
        """
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(tuple(self._cparts))
            return self._hash



# -*- coding: utf-8 -*-

"""
The ``S3Path`` public API class.
"""

# The import order is very important. The later one depends on the earlier one.

from .base import BaseS3Path
from .is_test import IsTestAPIMixin
from .uri import UriAPIMixin
from .relative import RelativePathAPIMixin
from .comparison import ComparisonAPIMixin
from .attribute import AttributeAPIMixin
from .joinpath import JoinPathAPIMixin
from .mutate import MutateAPIMixin
from .metadata import MetadataAPIMixin
from .bucket import BucketAPIMixin
from .tagging import TaggingAPIMixin
from .iter_objects import IterObjectsAPIMixin
from .exists import ExistsAPIMixin
from .rw import ReadAndWriteAPIMixin
from .delete import DeleteAPIMixin
from .upload import UploadAPIMixin
from .copy import CopyAPIMixin
from .sync import SyncAPIMixin
from .serde import SerdeAPIMixin
from .opener import OpenerAPIMixin


class S3Path(
    BaseS3Path,
    IsTestAPIMixin,
    UriAPIMixin,
    RelativePathAPIMixin,
    ComparisonAPIMixin,
    AttributeAPIMixin,
    JoinPathAPIMixin,
    MutateAPIMixin,
    MetadataAPIMixin,
    BucketAPIMixin,
    TaggingAPIMixin,
    IterObjectsAPIMixin,
    ExistsAPIMixin,
    ReadAndWriteAPIMixin,
    DeleteAPIMixin,
    UploadAPIMixin,
    CopyAPIMixin,
    SyncAPIMixin,
    SerdeAPIMixin,
    OpenerAPIMixin,
):
    """
    The ``S3Path`` public API class.
    """
